# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os
import time
import arrow
from datetime import datetime, timedelta

from superdesk.io import registered_feeding_services
from superdesk.io.commands.update_ingest import LAST_ITEM_UPDATE
import superdesk
import superdesk.tests as tests
from behave import given, when, then  # @UnresolvedImport
from flask import json
from eve.methods.common import parse
from superdesk import default_user_preferences, get_resource_service, utc, etree
from superdesk.io.feed_parsers import XMLFeedParser
from superdesk.utc import utcnow, get_expiry_date
from eve.io.mongo import MongoJSONEncoder
from base64 import b64encode
from wooper.general import fail_and_print_body, apply_path, \
    parse_json_response, WooperAssertionError
from wooper.expect import (
    expect_status, expect_status_in,
    expect_json, expect_json_length,
    expect_json_contains, expect_json_not_contains,
    expect_headers_contain,
)
from wooper.assertions import (
    assert_in, assert_equal)
from urllib.parse import urlparse
from os.path import basename
from superdesk.tests import test_user, get_prefixed_url, set_placeholder
from re import findall
from eve.utils import ParsedRequest
import shutil
from apps.dictionaries.resource import DICTIONARY_FILE
from test_factory import setup_auth_user


external_url = 'http://thumbs.dreamstime.com/z/digital-nature-10485007.jpg'
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S%z"


def test_json(context):
    try:
        response_data = json.loads(context.response.get_data())
    except Exception:
        fail_and_print_body(context.response, 'response is not valid json')
    context_data = json.loads(apply_placeholders(context, context.text))
    assert_equal(json_match(context_data, response_data), True,
                 msg=str(context_data) + '\n != \n' + str(response_data))
    return response_data


def test_json_with_string_field_value(context, field):
    try:
        response_data = json.loads(context.response.get_data())
    except Exception:
        fail_and_print_body(context.response, 'response is not valid json')
    context_data = json.loads(apply_placeholders(context, context.text))

    assert_equal(json_match(context_data[field], json.loads(response_data[field])), True,
                 msg=str(context_data) + '\n != \n' + str(response_data))
    return response_data


def test_key_is_present(key, context, response):
    """Test if given key is present in response.

    In case the context value is empty - "", {}, [] - it checks if it's non empty in response.

    If it's set in context to false, it will check that it's falsy/empty in response too.

    :param key
    :param context
    :param response
    """
    assert not isinstance(context[key], bool) or not response[key], \
        '"%s" should be empty or false, but it was "%s" in (%s)' % (key, response[key], response)


def assert_is_now(val, key):
    """Assert that given datetime value is now (with 5s tolerance).

    :param val: datetime
    :param key: val label - used for error reporting
    """
    now = arrow.get()
    val = arrow.get(val)
    assert val + timedelta(seconds=3) > now, '%s should be now, it is %s' % (key, val)


def json_match(context_data, response_data):
    if isinstance(context_data, dict):
        assert isinstance(response_data, dict), 'response data is not dict, but %s' % type(response_data)
        for key in context_data:
            if key not in response_data:
                print(key, ' not in ', response_data)
                return False
            if context_data[key] == "__any_value__":
                test_key_is_present(key, context_data, response_data)
                continue
            if context_data[key] == "__now__":
                assert_is_now(response_data[key], key)
                return True
            if not json_match(context_data[key], response_data[key]):
                return False
        return True
    elif isinstance(context_data, list):
        for item_context in context_data:
            found = False
            for item_response in response_data:
                if json_match(item_context, item_response):
                    found = True
                    break
            if not found:
                print(item_context, ' not in ', response_data)
                return False
        return True
    elif not isinstance(context_data, dict):
        if context_data != response_data:
            print('---' + str(context_data) + '---\n', ' != \n', '---' + str(response_data) + '---\n')
        return context_data == response_data


def get_fixture_path(context, fixture):
    path = context.app.settings.get('BEHAVE_TESTS_FIXTURES_PATH',
                                    os.path.join(os.path.abspath(os.path.dirname(__file__)), 'fixtures'))
    return os.path.join(path, fixture)


def get_macro_path(macro):
    abspath = os.path.abspath("macros")
    return os.path.join(abspath, macro)


def get_self_href(resource, context):
    assert '_links' in resource, 'expted "_links", but got only %s' % (resource)
    return resource['_links']['self']['href']


def get_res(url, context):
    response = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
    expect_status(response, 200)
    return json.loads(response.get_data())


def parse_date(datestr):
    return datetime.strptime(datestr, DATETIME_FORMAT)


def format_date(date_to_format):
    return date_to_format.strftime(DATETIME_FORMAT)


def assert_200(response):
    """Assert we get status code 200."""
    expect_status_in(response, (200, 201, 204))


def assert_404(response):
    """Assert we get status code 404."""
    assert response.status_code == 404, 'Expected 404, got %d' % (response.status_code)


def assert_ok(response):
    """Assert we get ok status within api response."""
    expect_status_in(response, (200, 201))
    expect_json_contains(response, {'_status': 'OK'})


def get_json_data(response):
    return json.loads(response.get_data())


def get_it(context):
    it = context.data[0]
    res = get_res('/%s/%s' % (context.resource, it['_id']), context)
    return get_self_href(res, context), res.get('_etag')


def if_match(context, etag):
    headers = []
    if etag:
        headers = [('If-Match', etag)]
    headers = unique_headers(headers, context.headers)
    return headers


def unique_headers(headers_to_add, old_headers):
    headers = dict(old_headers)
    for item in headers_to_add:
        headers.update({item[0]: item[1]})
    unique_headers = [(k, v) for k, v in headers.items()]
    return unique_headers


def patch_current_user(context, data):
    response = context.client.get(get_prefixed_url(context.app, '/users/%s' % context.user['_id']),
                                  headers=context.headers)
    user = json.loads(response.get_data())
    headers = if_match(context, user.get('_etag'))
    response = context.client.patch(get_prefixed_url(context.app, '/users/%s' % context.user['_id']),
                                    data=data, headers=headers)
    assert_ok(response)
    return response


def apply_placeholders(context, text):
    placeholders = getattr(context, 'placeholders', {})
    for placeholder in findall('#([^#"]+)#', text):
        if placeholder.startswith('DATE'):
            value = utcnow()
            unit = placeholder.find('+')
            if unit != -1:
                value += timedelta(days=int(placeholder[unit + 1]))
            else:
                unit = placeholder.find('-')
                if unit != -1:
                    value -= timedelta(days=int(placeholder[unit + 1]))

            value = format_date(value)
        elif placeholder not in placeholders:
            try:
                resource_name, field_name = placeholder.split('.', maxsplit=1)
            except:
                continue
            resource = getattr(context, resource_name, None)
            for name in field_name.split('.'):
                if not resource:
                    break

                resource = resource.get(name, None)

            if not resource:
                continue

            if isinstance(resource, datetime):
                value = format_date(resource)
            else:
                value = str(resource)
        else:
            value = placeholders[placeholder]
        text = text.replace('#%s#' % placeholder, value)
    return text


def get_resource_name(url):
    parsed_url = urlparse(url)
    return basename(parsed_url.path)


@given('empty "{resource}"')
def step_impl_given_empty(context, resource):
    if not is_user_resource(resource):
        with context.app.test_request_context(context.app.config['URL_PREFIX']):
            get_resource_service(resource).delete_action()


@given('"{resource}"')
def step_impl_given_(context, resource):
    data = apply_placeholders(context, context.text)
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        if not is_user_resource(resource):
            get_resource_service(resource).delete_action()

        items = [parse(item, resource) for item in json.loads(data)]
        if is_user_resource(resource):
            for item in items:
                item.setdefault('needs_activation', False)

        get_resource_service(resource).post(items)
        context.data = items
        context.resource = resource
        setattr(context, resource, items[-1])


@given('the "{resource}"')
def step_impl_given_the(context, resource):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        if not is_user_resource(resource):
            get_resource_service(resource).delete_action()

        orig_items = {}
        items = [parse(item, resource) for item in json.loads(context.text)]
        get_resource_service(resource).post(items)
        context.data = orig_items or items
        context.resource = resource


@given('ingest from "{provider}"')
def step_impl_given_resource_with_provider(context, provider):
    resource = 'ingest'
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        get_resource_service(resource).delete_action()
        items = [parse(item, resource) for item in json.loads(context.text)]
        for item in items:
            item['ingest_provider'] = context.providers[provider]
        get_resource_service(resource).post(items)
        context.data = items
        context.resource = resource


@given('config')
def step_impl_given_config(context):
    tests.setup(context, json.loads(context.text))
    setup_auth_user(context)


@given('we have "{role_name}" role')
def step_impl_given_role(context, role_name):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        role = get_resource_service('roles').find_one(name=role_name, req=None)
        data = MongoJSONEncoder().encode({'role': role.get('_id')})
    response = patch_current_user(context, data)
    assert_ok(response)


@given('we have "{user_type}" as type of user')
def step_impl_given_user_type(context, user_type):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        data = json.dumps({'user_type': user_type})
    response = patch_current_user(context, data)
    assert_ok(response)


@when('we post to auth')
def step_impl_when_auth(context):
    data = context.text
    context.response = context.client.post(get_prefixed_url(context.app, '/auth'), data=data, headers=context.headers)
    if context.response.status_code == 200 or context.response.status_code == 201:
        item = json.loads(context.response.get_data())
        if item.get('_id'):
            set_placeholder(context, 'AUTH_ID', item['_id'])
        context.headers.append(('Authorization', b'basic ' + b64encode(item['token'].encode('ascii') + b':')))
        context.user = item['user']


@given('we create a new macro "{macro_name}"')
def step_create_new_macro(context, macro_name):
    src = get_fixture_path(context, macro_name)
    dst = get_macro_path(macro_name)
    shutil.copyfile(src, dst)


@when('we fetch from "{provider_name}" ingest "{guid}"')
def step_impl_fetch_from_provider_ingest(context, provider_name, guid):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        fetch_from_provider(context, provider_name, guid)


def embed_routing_scheme_rules(scheme):
    """Fetch all content filters referenced by the given routing scheme and
    embed them into the latter (replacing the plain references to filters).

    :param dict scheme: routing scheme configuration
    """
    filters_service = superdesk.get_resource_service('content_filters')

    rules_filters = (
        (rule, str(rule['filter']))
        for rule in scheme['rules'] if rule.get('filter'))

    for rule, filter_id in rules_filters:
        content_filter = filters_service.find_one(_id=filter_id, req=None)
        rule['filter'] = content_filter


@when('we fetch from "{provider_name}" ingest "{guid}" using routing_scheme')
def step_impl_fetch_from_provider_ingest_using_routing(context, provider_name, guid):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        _id = apply_placeholders(context, context.text)
        routing_scheme = get_resource_service('routing_schemes').find_one(_id=_id, req=None)
        embed_routing_scheme_rules(routing_scheme)
        fetch_from_provider(context, provider_name, guid, routing_scheme)


@when('we ingest and fetch "{provider_name}" "{guid}" to desk "{desk}" stage "{stage}" using routing_scheme')
def step_impl_fetch_from_provider_ingest_using_routing_with_desk(context, provider_name, guid, desk, stage):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        _id = apply_placeholders(context, context.text)
        desk_id = apply_placeholders(context, desk)
        stage_id = apply_placeholders(context, stage)
        routing_scheme = get_resource_service('routing_schemes').find_one(_id=_id, req=None)
        embed_routing_scheme_rules(routing_scheme)
        fetch_from_provider(context, provider_name, guid, routing_scheme, desk_id, stage_id)


@when('we ingest with routing scheme "{provider_name}" "{guid}"')
def step_impl_ingest_with_routing_scheme(context, provider_name, guid):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        _id = apply_placeholders(context, context.text)
        routing_scheme = get_resource_service('routing_schemes').find_one(_id=_id, req=None)
        embed_routing_scheme_rules(routing_scheme)
        fetch_from_provider(context, provider_name, guid, routing_scheme)


def fetch_from_provider(context, provider_name, guid, routing_scheme=None, desk_id=None, stage_id=None):
    ingest_provider_service = get_resource_service('ingest_providers')
    provider = ingest_provider_service.find_one(name=provider_name, req=None)
    provider['routing_scheme'] = routing_scheme
    if 'rule_set' in provider:
        rule_set = get_resource_service('rule_sets').find_one(_id=provider['rule_set'], req=None)
    else:
        rule_set = None

    provider_service = registered_feeding_services[provider['feeding_service']]
    provider_service = provider_service.__class__()

    if provider.get('name', '').lower() in ('aap', 'teletype', 'dpa'):
        file_path = os.path.join(provider.get('config', {}).get('path', ''), guid)
        feeding_parser = provider_service.get_feed_parser(provider)
        if isinstance(feeding_parser, XMLFeedParser):
            with open(file_path, 'r') as f:
                xml_string = etree.etree.fromstring(f.read())
                items = [feeding_parser.parse(xml_string, provider)]
        else:
            items = [feeding_parser.parse(file_path, provider)]
    else:
        provider_service.provider = provider
        provider_service.URL = provider.get('config', {}).get('url')
        items = provider_service.fetch_ingest(guid)

    for item in items:
        item['versioncreated'] = utcnow()
        item['expiry'] = utcnow() + timedelta(minutes=20)

        if desk_id:
            from bson.objectid import ObjectId
            item['task'] = {'desk': ObjectId(desk_id), 'stage': ObjectId(stage_id)}

    failed = context.ingest_items(items, provider, provider_service, rule_set=rule_set,
                                  routing_scheme=provider.get('routing_scheme'))
    assert len(failed) == 0, failed

    provider = ingest_provider_service.find_one(name=provider_name, req=None)
    ingest_provider_service.system_update(provider['_id'], {LAST_ITEM_UPDATE: utcnow()}, provider)

    for item in items:
        set_placeholder(context, '{}.{}'.format(provider_name, item['guid']), item['_id'])


@when('we post to "{url}"')
def step_impl_when_post_url(context, url):
    post_data(context, url)


@when('we post to "{url}" with delay')
def step_impl_when_post_url(context, url):
    time.sleep(1)
    post_data(context, url)


def set_user_default(url, data):
    if is_user_resource(url):
        user = json.loads(data)
        user.setdefault('needs_activation', False)
        data = json.dumps(user)


def get_response_etag(response):
    return json.loads(response.get_data())['_etag']


@when('we save etag')
def step_when_we_save_etag(context):
    context.etag = get_response_etag(context.response)


@then('we get same etag')
def step_then_we_get_same_etag(context):
    assert context.etag == get_response_etag(context.response), 'etags not matching'


def store_placeholder(context, url):
    if context.response.status_code in (200, 201):
        item = json.loads(context.response.get_data())
        if item['_status'] == 'OK' and item.get('_id'):
            setattr(context, get_resource_name(url), item)


def post_data(context, url, success=False):
    with context.app.mail.record_messages() as outbox:
        data = apply_placeholders(context, context.text)
        url = apply_placeholders(context, url)
        set_user_default(url, data)
        context.response = context.client.post(get_prefixed_url(context.app, url),
                                               data=data, headers=context.headers)
        if success:
            assert_ok(context.response)

        item = json.loads(context.response.get_data())
        context.outbox = outbox
        store_placeholder(context, url)
        return item


@when('we post to "{url}" with "{tag}" and success')
def step_impl_when_post_url_with_tag(context, url, tag):
    item = post_data(context, url, True)
    if item.get('_id'):
        set_placeholder(context, tag, item.get('_id'))


@given('we have "{url}" with "{tag}" and success')
def step_impl_given_post_url_with_tag(context, url, tag):
    item = post_data(context, url, True)
    if item.get('_id'):
        set_placeholder(context, tag, item.get('_id'))


@when('we post to "{url}" with success')
def step_impl_when_post_url_with_success(context, url):
    post_data(context, url, True)


@when('we put to "{url}"')
def step_impl_when_put_url(context, url):
    with context.app.mail.record_messages() as outbox:
        data = apply_placeholders(context, context.text)
        href = get_self_href(url)
        context.response = context.client.put(get_prefixed_url(context.app, href), data=data, headers=context.headers)
        assert_ok(context.response)
        context.outbox = outbox


@when('we get "{url}"')
def when_we_get_url(context, url):
    url = apply_placeholders(context, url).encode('ascii').decode('unicode-escape')
    headers = []
    if context.text:
        for line in context.text.split('\n'):
            key, val = line.split(': ')
            headers.append((key, val))
    headers = unique_headers(headers, context.headers)
    url = apply_placeholders(context, url)
    context.response = context.client.get(get_prefixed_url(context.app, url), headers=headers)


@when('we get dictionary "{dictionary_id}"')
def when_we_get_dictionary(context, dictionary_id):
    dictionary_id = apply_placeholders(context, dictionary_id)
    url = '/dictionaries/' + dictionary_id + '?projection={"content": 1}'
    return when_we_get_url(context, url)


@then('we get latest')
def step_impl_we_get_latest(context):
    data = get_json_data(context.response)
    href = get_self_href(data, context)
    headers = if_match(context, data.get('_etag'))
    href = get_prefixed_url(context.app, href)
    context.response = context.client.get(href, headers=headers)
    assert_200(context.response)


@when('we find for "{resource}" the id as "{name}" by "{search_criteria}"')
def when_we_find_for_resource_the_id_as_name_by_search_criteria(context, resource, name, search_criteria):
    url = '/' + resource + '?where=' + search_criteria
    context.response = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
    if context.response.status_code == 200:
        expect_json_length(context.response, 1, path='_items')
        item = json.loads(context.response.get_data())
        item = item['_items'][0]
        if item.get('_id'):
            set_placeholder(context, name, item['_id'])


@when('we delete "{url}"')
def step_impl_when_delete_url(context, url):
    with context.app.mail.record_messages() as outbox:
        url = apply_placeholders(context, url)
        res = get_res(url, context)
        href = get_self_href(res, context)
        headers = if_match(context, res.get('_etag'))
        href = get_prefixed_url(context.app, href)
        context.response = context.client.delete(href, headers=headers)
        context.outbox = outbox


@when('we delete all sessions "{url}"')
def step_impl_when_delete_all_url(context, url):
    with context.app.mail.record_messages() as outbox:
        url = apply_placeholders(context, url)
        headers = context.headers
        href = get_prefixed_url(context.app, url)
        context.response = context.client.delete(href, headers=headers)
        context.outbox = outbox


@when('we delete latest')
def when_we_delete_it(context):
    with context.app.mail.record_messages() as outbox:
        res = get_json_data(context.response)
        href = get_self_href(res, context)
        headers = if_match(context, res.get('_etag'))
        href = get_prefixed_url(context.app, href)
        context.response = context.client.delete(href, headers=headers)
        context.email = outbox


@when('we patch "{url}"')
def step_impl_when_patch_url(context, url):
    with context.app.mail.record_messages() as outbox:
        url = apply_placeholders(context, url)
        res = get_res(url, context)
        href = get_self_href(res, context)
        headers = if_match(context, res.get('_etag'))
        data = apply_placeholders(context, context.text)
        href = get_prefixed_url(context.app, href)
        context.response = context.client.patch(href, data=data, headers=headers)
        context.outbox = outbox


@when('we patch latest')
def step_impl_when_patch_again(context):
    with context.app.mail.record_messages() as outbox:
        data = get_json_data(context.response)
        href = get_prefixed_url(context.app, get_self_href(data, context))
        headers = if_match(context, data.get('_etag'))
        data2 = apply_placeholders(context, context.text)
        context.response = context.client.patch(href, data=data2, headers=headers)
        if context.response.status_code in (200, 201):
            item = json.loads(context.response.get_data())
            if item['_status'] == 'OK' and item.get('_id'):
                setattr(context, get_resource_name(href), item)
        assert_ok(context.response)
        context.outbox = outbox


@when('we patch latest without assert')
def step_impl_when_patch_without_assert(context):
    data = get_json_data(context.response)
    href = get_prefixed_url(context.app, get_self_href(data, context))
    headers = if_match(context, data.get('_etag'))
    data2 = apply_placeholders(context, context.text)
    context.response = context.client.patch(href, data=data2, headers=headers)


@when('we patch routing scheme "{url}"')
def step_impl_when_patch_routing_scheme(context, url):
    with context.app.mail.record_messages() as outbox:
        url = apply_placeholders(context, url)
        res = get_res(url, context)
        href = get_self_href(res, context)
        headers = if_match(context, res.get('_etag'))
        data = json.loads(apply_placeholders(context, context.text))
        res.get('rules', []).append(data)
        context.response = context.client.patch(get_prefixed_url(context.app, href),
                                                data=json.dumps({'rules': res.get('rules', [])}),
                                                headers=headers)
        context.outbox = outbox


@when('we patch given')
def step_impl_when_patch(context):
    with context.app.mail.record_messages() as outbox:
        href, etag = get_it(context)
        headers = if_match(context, etag)
        context.response = context.client.patch(get_prefixed_url(context.app, href), data=context.text, headers=headers)
        assert_ok(context.response)
        context.outbox = outbox


@when('we get given')
def step_impl_when_get(context):
    href, _etag = get_it(context)
    context.response = context.client.get(get_prefixed_url(context.app, href), headers=context.headers)


@when('we restore version {version}')
def step_impl_when_restore_version(context, version):
    data = get_json_data(context.response)
    href = get_self_href(data, context)
    headers = if_match(context, data.get('_etag'))
    text = '{"type": "text", "old_version": %s, "last_version": %s}' % (version, data.get('_current_version'))
    context.response = context.client.put(get_prefixed_url(context.app, href), data=text, headers=headers)
    assert_ok(context.response)


@when('we upload a file "{filename}" to "{dest}"')
def step_impl_when_upload_image(context, filename, dest):
    upload_file(context, dest, filename, 'media')


@when('we upload a binary file with cropping')
def step_impl_when_upload_with_crop(context):
    data = {'CropTop': '0', 'CropLeft': '0', 'CropBottom': '333', 'CropRight': '333'}
    upload_file(context, '/upload', 'bike.jpg', 'media', data)


@when('upload a file "{file_name}" to "{destination}" with "{guid}"')
def step_impl_when_upload_image_with_guid(context, file_name, destination, guid):
    upload_file(context, destination, file_name, 'media', {'guid': guid})


@when('we upload a new dictionary with success')
def when_upload_dictionary(context):
    data = json.loads(apply_placeholders(context, context.text))
    upload_file(context, '/dictionaries', 'test_dict.txt', DICTIONARY_FILE, data)
    assert_ok(context.response)


@when('we upload to an existing dictionary with success')
def when_upload_patch_dictionary(context):
    data = json.loads(apply_placeholders(context, context.text))
    url = apply_placeholders(context, '/dictionaries/#dictionaries._id#')
    etag = apply_placeholders(context, '#dictionaries._etag#')
    upload_file(context, url, 'test_dict2.txt', DICTIONARY_FILE, data, 'patch', [('If-Match', etag)])
    assert_ok(context.response)


def upload_file(context, dest, filename, file_field, extra_data=None, method='post', user_headers=[]):
    with open(get_fixture_path(context, filename), 'rb') as f:
        data = {file_field: f}
        if extra_data:
            data.update(extra_data)
        headers = [('Content-Type', 'multipart/form-data')]
        headers.extend(user_headers)
        headers = unique_headers(headers, context.headers)
        url = get_prefixed_url(context.app, dest)
        context.response = getattr(context.client, method)(url, data=data, headers=headers)
        assert_ok(context.response)
        store_placeholder(context, url)


@when('we upload a file from URL')
def step_impl_when_upload_from_url(context):
    data = {'URL': external_url}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/upload'), data=data, headers=headers)


@when('we upload a file from URL with cropping')
def step_impl_when_upload_from_url_with_crop(context):
    data = {'URL': external_url,
            'CropTop': '0',
            'CropLeft': '0',
            'CropBottom': '333',
            'CropRight': '333'}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/upload'), data=data, headers=headers)


@when('we get user profile')
def step_impl_when_get_user(context):
    profile_url = '/%s/%s' % ('users', context.user['_id'])
    context.response = context.client.get(get_prefixed_url(context.app, profile_url), headers=context.headers)


@then('we get new resource')
def step_impl_then_get_new(context):
    assert_ok(context.response)
    expect_json_contains(context.response, 'self', path='_links')
    if context.text is not None:
        return test_json(context)


@then('we get next take as "{new_take}"')
def step_impl_then_get_next_take(context, new_take):
    step_impl_we_get_latest(context)
    data = get_json_data(context.response)
    set_placeholder(context, new_take, data['_id'])
    set_placeholder(context, 'TAKE_PACKAGE', data['takes']['_id'])
    test_json(context)


@then('we get error {code}')
def step_impl_then_get_error(context, code):
    expect_status(context.response, int(code))
    if context.text:
        test_json(context)


@then('we get list with {total_count} items')
def step_impl_then_get_list(context, total_count):
    assert_200(context.response)
    data = get_json_data(context.response)
    int_count = int(total_count.replace('+', ''))

    if '+' in total_count:
        assert int_count <= data['_meta']['total'], '%d items is not enough' % data['_meta']['total']
    else:
        assert int_count == data['_meta']['total'], 'got %d' % (data['_meta']['total'])
    if context.text:
        test_json(context)


@then('we get "{value}" in formatted output')
def step_impl_then_get_formatted_output(context, value):
    assert_200(context.response)
    value = apply_placeholders(context, value)
    data = get_json_data(context.response)
    for item in data['_items']:
        if value in item['formatted_item']:
            return
    assert False


@then('we get "{value}" in formatted output as "{group}" story for subscriber "{sub}"')
def step_impl_then_get_formatted_output(context, value, group, sub):
    assert_200(context.response)
    value = apply_placeholders(context, value)
    data = get_json_data(context.response)
    for item in data['_items']:
        if item['subscriber_id'] != sub:
            continue

        try:
            formatted_data = json.loads(item['formatted_item'])
        except:
            continue

        for group_item in formatted_data.get('associations', {}).get(group, []):
            if group_item.get('_id', '') == value:
                return
    assert False


@then('we get "{value}" as "{group}" story for subscriber "{sub}" in package "{pck}"')
def step_impl_then_get_formatted_output_pck(context, value, group, sub, pck):
    assert_200(context.response)
    value = apply_placeholders(context, value)
    data = get_json_data(context.response)
    for item in data['_items']:
        if item['item_id'] != pck:
            continue

        if item['subscriber_id'] != sub:
            continue

        try:
            formatted_data = json.loads(item['formatted_item'])
        except:
            continue

        for group_item in formatted_data.get('associations', {}).get(group, []):
            if group_item.get('_id', '') == value:
                return
    assert False


@then('we get "{value}" as "{group}" story for subscriber "{sub}" not in package "{pck}" version "{v}"')
def step_impl_then_get_formatted_output_pck(context, value, group, sub, pck, v):
    assert_200(context.response)
    value = apply_placeholders(context, value)
    data = get_json_data(context.response)
    for item in data['_items']:
        if item['item_id'] == pck:
            if item['subscriber_id'] == sub and str(item['item_version']) == v:
                try:
                    formatted_data = json.loads(item['formatted_item'])
                except:
                    continue

                for group_item in formatted_data.get('associations', {}).get(group, []):
                    if group_item.get('_id', '') == value:
                        assert False

                assert True
                return
    assert False


@then('we get "{value}" in formatted output as "{group}" newsml12 story')
def step_impl_then_get_formatted_output_newsml(context, value, group):
    assert_200(context.response)
    value = apply_placeholders(context, value)
    data = get_json_data(context.response)
    for item in data['_items']:
        if '<' + group + '>' + value + '</' + group + '>' in item['formatted_item']:
            return
    assert False


@then('we get no "{field}"')
def step_impl_then_get_nofield(context, field):
    assert_200(context.response)
    expect_json_not_contains(context.response, field)


@then('expect json in "{path}"')
def step_impl_then_get_nofield_in_path(context, path):
    assert_200(context.response)
    expect_json(context.response, context.text, path)


@then('we get existing resource')
def step_impl_then_get_existing(context):
    assert_200(context.response)
    test_json(context)


@then('we get existing saved search')
def step_impl_then_get_existing_saved_search(context):
    assert_200(context.response)
    test_json_with_string_field_value(context, 'filter')


@then('we get OK response')
def step_impl_then_get_ok(context):
    assert_200(context.response)


@then('we get response code {code}')
def step_impl_then_get_code(context, code):
    expect_status(context.response, int(code))


@then('we get updated response')
def step_impl_then_get_updated(context):
    assert_ok(context.response)
    if context.text:
        test_json(context)


@then('we get "{key}" in "{url}"')
def step_impl_then_get_key_in_url(context, key, url):
    url = apply_placeholders(context, url)
    res = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
    assert_200(res)
    expect_json_contains(res, key)


@then('we get file metadata')
def step_impl_then_get_file_meta(context):
    assert len(
        apply_path(
            parse_json_response(context.response),
            'filemeta'
        ).items()
    ) > 0
    'expected non empty metadata dictionary'


@then('we get "{filename}" metadata')
def step_impl_then_get_given_file_meta(context, filename):
    if filename == 'bike.jpg':
        metadata = {
            'ycbcrpositioning': 1,
            'imagelength': 2448,
            'exifimagewidth': 2448,
            'meteringmode': 2,
            'datetimedigitized': '2013:08:01 16:19:28',
            'exposuremode': 0,
            'flashpixversion': '0100',
            'isospeedratings': 80,
            'length': 469900,
            'imageuniqueid': 'f3533c05daef2debe6257fd99e058eec',
            'datetimeoriginal': '2013:08:01 16:19:28',
            'whitebalance': 0,
            'exposureprogram': 3,
            'colorspace': 1,
            'exifimageheight': 3264,
            'software': 'Google',
            'resolutionunit': 2,
            'make': 'SAMSUNG',
            'maxaperturevalue': [276, 100],
            'aperturevalue': [276, 100],
            'scenecapturetype': 0,
            'exposuretime': [1, 2004],
            'datetime': '2013:08:01 16:19:28',
            'exifoffset': 216,
            'yresolution': [72, 1],
            'orientation': 1,
            'componentsconfiguration': '0000',
            'exifversion': '0220',
            'focallength': [37, 10],
            'flash': 0,
            'model': 'GT-I9300',
            'xresolution': [72, 1],
            'fnumber': [26, 10],
            'imagewidth': 3264
        }
    elif filename == 'green.ogg':
        metadata = {
            'producer': 'Lavf54.59.103',
            'music_genre': 'New Age',
            'sample_rate': '44100',
            'artist': 'Maxime Abbey',
            'length': 368058,
            'bit_rate': '160000',
            'title': 'Green Hills',
            'mime_type': 'audio/vorbis',
            'format_version': 'Vorbis version 0',
            'compression': 'Vorbis',
            'duration': '0:00:20.088163',
            'endian': 'Little endian',
            'nb_channel': '2'
        }
    elif filename == 'this_week_nasa.mp4':
        metadata = {
            'mime_type': 'video/mp4',
            'creation_date': '1904-01-01T00:00:00+00:00',
            'duration': '0:00:10.224000',
            'width': '480',
            'length': 877869,
            'comment': 'User volume: 100.0%',
            'height': '270',
            'endian': 'Big endian',
            'last_modification': '1904-01-01T00:00:00+00:00'
        }
    else:
        raise NotImplementedError("No metadata for file '{}'.".format(filename))

    expect_json(
        context.response,
        metadata,
        path='filemeta'
    )


@then('we get "{type}" renditions')
def step_impl_then_get_renditions(context, type):
    expect_json_contains(context.response, 'renditions')
    renditions = apply_path(parse_json_response(context.response), 'renditions')
    assert isinstance(renditions, dict), 'expected dict for image renditions'
    for rend_name in context.app.config['RENDITIONS'][type]:
        desc = renditions[rend_name]
        assert isinstance(desc, dict), 'expected dict for rendition description'
        assert 'href' in desc, 'expected href in rendition description'
        assert 'media' in desc, 'expected media identifier in rendition description'
        we_can_fetch_a_file(context, desc['href'], 'image/jpeg')


@then('item "{item_id}" is unlocked')
def then_item_is_unlocked(context, item_id):
    assert_200(context.response)
    data = json.loads(context.response.get_data())
    assert data.get('lock_user', None) is None, 'item is locked by user #{0}'.format(data.get('lock_user'))


@then('item "{item_id}" is locked')
def then_item_is_locked(context, item_id):
    assert_200(context.response)
    resp = parse_json_response(context.response)
    assert resp['lock_user'] is not None


@then('item "{item_id}" is assigned')
def then_item_is_assigned(context, item_id):
    resp = parse_json_response(context.response)
    assert resp['task'].get('user', None) is not None, 'item is not assigned'


@then('we get rendition "{name}" with mimetype "{mimetype}"')
def step_impl_then_get_rendition_with_mimetype(context, name, mimetype):
    expect_json_contains(context.response, 'renditions')
    renditions = apply_path(parse_json_response(context.response), 'renditions')
    assert isinstance(renditions, dict), 'expected dict for image renditions'
    desc = renditions[name]
    assert isinstance(desc, dict), 'expected dict for rendition description'
    assert 'href' in desc, 'expected href in rendition description'
    we_can_fetch_a_file(context, desc['href'], mimetype)
    set_placeholder(context, "rendition.{}.href".format(name), desc['href'])


@when('we get updated media from archive')
def get_updated_media_from_archive(context):
    url = 'archive/%s' % context._id
    when_we_get_url(context, url)
    assert_200(context.response)


@then('baseImage rendition is updated')
def check_base_image_rendition(context):
    check_rendition(context, 'baseImage')


@then('original rendition is updated with link to file having mimetype "{mimetype}"')
def check_original_rendition(context, mimetype):
    rv = parse_json_response(context.response)
    link_to_file = rv['renditions']['original']['href']
    assert link_to_file
    we_can_fetch_a_file(context, link_to_file, mimetype)


@then('thumbnail rendition is updated')
def check_thumbnail_rendition(context):
    check_rendition(context, 'thumbnail')


def check_rendition(context, rendition_name):
    rv = parse_json_response(context.response)
    assert rv['renditions'][rendition_name] != context.renditions[rendition_name], rv['renditions']


@then('we get "{key}"')
def step_impl_then_get_key(context, key):
    assert_200(context.response)
    expect_json_contains(context.response, key)
    item = json.loads(context.response.get_data())
    set_placeholder(context, '%s' % key, item[key])


@then('we get action in user activity')
def step_impl_then_get_action(context):
    response = context.client.get(get_prefixed_url(context.app, '/activity'), headers=context.headers)
    expect_json_contains(response, '_items')


@then('we get a file reference')
def step_impl_then_get_file(context):
    assert_200(context.response)
    expect_json_contains(context.response, 'renditions')
    data = get_json_data(context.response)
    url = '/upload/%s' % data['_id']
    headers = [('Accept', 'application/json')]
    headers = unique_headers(headers, context.headers)
    response = context.client.get(get_prefixed_url(context.app, url), headers=headers)
    assert_200(response)
    assert len(response.get_data()), response
    assert response.mimetype == 'application/json', response.mimetype
    expect_json_contains(response, 'renditions')
    expect_json_contains(response, {'mimetype': 'image/jpeg'})
    fetched_data = get_json_data(context.response)
    context.fetched_data = fetched_data


@then('we get cropped data smaller than "{max_size}"')
def step_impl_then_get_cropped_file(context, max_size):
    assert int(context.fetched_data['filemeta']['length']) < int(max_size), 'was expecting smaller image'


@then('we can fetch a data_uri')
def step_impl_we_fetch_data_uri(context):
    we_can_fetch_a_file(context, context.fetched_data['renditions']['original']['href'], 'image/jpeg')


@then('we fetch a file "{url}"')
def step_impl_we_cannot_fetch_file(context, url):
    url = apply_placeholders(context, url)
    headers = [('Accept', 'application/json')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.get(get_prefixed_url(context.app, url), headers=headers)


def we_can_fetch_a_file(context, url, mimetype):
    headers = [('Accept', 'application/json')]
    headers = unique_headers(headers, context.headers)
    response = context.client.get(get_prefixed_url(context.app, url), headers=headers)
    assert_200(response)
    assert len(response.get_data()), response
    assert response.mimetype == mimetype, response.mimetype


@then('we can delete that file')
def step_impl_we_delete_file(context):
    url = '/upload/%s' % context.fetched_data['_id']
    context.headers.append(('Accept', 'application/json'))
    headers = if_match(context, context.fetched_data.get('_etag'))
    response = context.client.delete(get_prefixed_url(context.app, url), headers=headers)
    assert_200(response)
    response = context.client.get(get_prefixed_url(context.app, url), headers=headers)
    assert_404(response)


@then('we get a picture url')
def step_impl_then_get_picture(context):
    assert_ok(context.response)
    expect_json_contains(context.response, 'picture_url')


@then('we get aggregations "{keys}"')
def step_impl_then_get_aggs(context, keys):
    assert_200(context.response)
    expect_json_contains(context.response, '_aggregations')
    data = get_json_data(context.response)
    aggs = data['_aggregations']
    for key in keys.split(','):
        assert_in(key, aggs)


@then('the file is stored localy')
def step_impl_then_file(context):
    assert_200(context.response)
    folder = context.app.config['UPLOAD_FOLDER']
    assert os.path.exists(os.path.join(folder, context.filename))


@then('we get version {version}')
def step_impl_then_get_version(context, version):
    assert_200(context.response)
    expect_json_contains(context.response, {'_current_version': int(version)})


@then('the field "{field}" value is "{value}"')
def step_impl_then_get_field_value(context, field, value):
    assert_200(context.response)
    expect_json_contains(context.response, {field: value})


@then('we get etag matching "{url}"')
def step_impl_then_get_etag(context, url):
    if context.app.config['IF_MATCH']:
        assert_200(context.response)
        expect_json_contains(context.response, '_etag')
        etag = get_json_data(context.response).get('_etag')
        response = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
        expect_json_contains(response, {'_etag': etag})


@then('we get not modified response')
def step_impl_then_not_modified(context):
    expect_status(context.response, 304)


@then('we get "{header}" header')
def step_impl_then_get_header(context, header):
    expect_headers_contain(context.response, header)


@then('we get link to "{resource}"')
def then_we_get_link_to_resource(context, resource):
    doc = get_json_data(context.response)
    self_link = doc.get('_links').get('self')
    assert resource in self_link['href'], 'expect link to "%s", got %s' % (resource, self_link)


@then('we get deleted response')
def then_we_get_deleted_response(context):
    assert_200(context.response)


@when('we post to reset_password we get email with token')
def we_post_to_reset_password(context):
    data = {'email': 'foo@bar.org'}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    with context.app.mail.record_messages() as outbox:
        context.response = context.client.post(get_prefixed_url(context.app, '/reset_user_password'),
                                               data=data, headers=headers)
        expect_status_in(context.response, (200, 201))
        assert len(outbox) == 1
        assert outbox[0].subject == "Reset password"
        email_text = outbox[0].body
        assert "24" in email_text
        words = email_text.split()
        url = urlparse(words[words.index("link") + 1])
        token = url.fragment.split('token=')[-1]
        assert token
        context.token = token


@then('we can check if token is valid')
def we_can_check_token_is_valid(context):
    data = {'token': context.token}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/reset_user_password'),
                                           data=data, headers=headers)
    expect_status_in(context.response, (200, 201))


@then('we update token to be expired')
def we_update_token_to_expired(context):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        expiry = utc.utcnow() - timedelta(days=2)
        reset_request = get_resource_service('reset_user_password').find_one(req=None, token=context.token)
        reset_request['expire_time'] = expiry
        id = reset_request.pop('_id')
        get_resource_service('reset_user_password').patch(id, reset_request)


@then('token is invalid')
def check_token_invalid(context):
    data = {'token': context.token}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/reset_user_password'),
                                           data=data, headers=headers)
    expect_status_in(context.response, (403, 401))


@when('we post to reset_password we do not get email with token')
def we_post_to_reset_password_it_fails(context):
    data = {'email': 'foo@bar.org'}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    with context.app.mail.record_messages() as outbox:
        context.response = context.client.post(get_prefixed_url(context.app, '/reset_user_password'),
                                               data=data, headers=headers)
        expect_status_in(context.response, (200, 201))
        assert len(outbox) == 0


def start_reset_password_for_user(context):
    data = {'token': context.token, 'password': 'test_pass'}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/reset_user_password'),
                                           data=data, headers=headers)


@then('we fail to reset password for user')
def we_fail_to_reset_password_for_user(context):
    start_reset_password_for_user(context)
    step_impl_then_get_error(context, 403)


@then('we reset password for user')
def we_reset_password_for_user(context):
    start_reset_password_for_user(context)
    expect_status_in(context.response, (200, 201))

    auth_data = {'username': 'foo', 'password': 'test_pass'}
    headers = [('Content-Type', 'multipart/form-data')]
    headers = unique_headers(headers, context.headers)
    context.response = context.client.post(get_prefixed_url(context.app, '/auth'), data=auth_data, headers=headers)
    expect_status_in(context.response, (200, 201))


@when('we switch user')
def when_we_switch_user(context):
    user = {'username': 'test-user-2', 'password': 'pwd', 'is_active': True,
            'needs_activation': False, 'sign_off': 'foo'}
    setup_auth_user(context, user)
    set_placeholder(context, 'USERS_ID', str(context.user['_id']))


@when('we setup test user')
def when_we_setup_test_user(context):
    setup_auth_user(context, test_user)


@when('we get my "{url}"')
def when_we_get_my_url(context, url):
    user_id = str(context.user.get('_id'))
    my_url = '{0}?where={1}'.format(url, json.dumps({'user': user_id}))
    return when_we_get_url(context, my_url)


@when('we get user "{resource}"')
def when_we_get_user_resource(context, resource):
    url = '/users/{0}/{1}'.format(str(context.user.get('_id')), resource)
    return when_we_get_url(context, url)


@then('we get embedded items')
def we_get_embedded_items(context):
    response_data = json.loads(context.response.get_data())
    href = get_self_href(response_data, context)
    url = href + '/?embedded={"items": 1}'
    context.response = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
    assert_200(context.response)
    context.response_data = json.loads(context.response.get_data())
    assert len(context.response_data['items']['view_items']) == 2


@when('we reset notifications')
def step_when_we_reset_notifications(context):
    context.app.notification_client.reset()


@then('we get notifications')
def then_we_get_notifications(context):
    notifications = context.app.notification_client.messages
    notifications_data = [json.loads(notification) for notification in notifications]
    context_data = json.loads(apply_placeholders(context, context.text))
    assert_equal(json_match(context_data, notifications_data), True,
                 msg=str(context_data) + '\n != \n' + str(notifications_data))


@then('we get default preferences')
def get_default_prefs(context):
    response_data = json.loads(context.response.get_data())
    assert_equal(response_data['user_preferences'], default_user_preferences)


@when('we spike "{item_id}"')
def step_impl_when_spike_url(context, item_id):
    item_id = apply_placeholders(context, item_id)
    res = get_res('/archive/' + item_id, context)
    headers = if_match(context, res.get('_etag'))

    context.response = context.client.patch(get_prefixed_url(context.app, '/archive/spike/' + item_id),
                                            data='{"state": "spiked"}', headers=headers)


@when('we spike fetched item')
def step_impl_when_spike_fetched_item(context):
    data = json.loads(apply_placeholders(context, context.text))
    item_id = data["_id"]
    res = get_res('/archive/' + item_id, context)
    headers = if_match(context, res.get('_etag'))

    context.response = context.client.patch(get_prefixed_url(context.app, '/archive/spike/' + item_id),
                                            data='{"state": "spiked"}', headers=headers)


@when('we unspike "{item_id}"')
def step_impl_when_unspike_url(context, item_id):
    item_id = apply_placeholders(context, item_id)
    res = get_res('/archive/' + item_id, context)
    headers = if_match(context, res.get('_etag'))
    context.response = context.client.patch(get_prefixed_url(context.app, '/archive/unspike/' + item_id),
                                            data='{}', headers=headers)


@then('we get spiked content "{item_id}"')
def get_spiked_content(context, item_id):
    item_id = apply_placeholders(context, item_id)
    url = 'archive/{0}'.format(item_id)
    when_we_get_url(context, url)
    assert_200(context.response)
    response_data = json.loads(context.response.get_data())
    assert_equal(response_data['state'], 'spiked')
    assert_equal(response_data['operation'], 'spike')


@then('we get unspiked content "{id}"')
def get_unspiked_content(context, id):
    url = 'archive/{0}'.format(id)
    when_we_get_url(context, url)
    assert_200(context.response)
    response_data = json.loads(context.response.get_data())
    assert_equal(response_data['state'], 'draft')
    assert_equal(response_data['operation'], 'unspike')
    # Tolga Akin (05/11/14)
    # Expiry value doesn't get set to None properly in Elastic.
    # Discussed with Petr so we'll look into this later
    # assert_equal(response_data['expiry'], None)


@then('we get global content expiry')
def get_global_content_expiry(context):
    validate_expired_content(context, context.app.config['CONTENT_EXPIRY_MINUTES'], utcnow())


@then('we get content expiry {minutes}')
def get_content_expiry(context, minutes):
    validate_expired_content(context, minutes, utcnow())


@then('we get expiry for schedule and embargo content {minutes} minutes after "{future_date}"')
def get_content_expiry(context, minutes, future_date):
    future_date = parse_date(apply_placeholders(context, future_date))
    validate_expired_content(context, minutes, future_date)


@then('we get desk spike expiry after "{test_minutes}"')
def get_desk_spike_expiry(context, test_minutes):
    validate_expired_content(context, test_minutes, utcnow())


def validate_expired_content(context, minutes, start_datetime):
    response_data = json.loads(context.response.get_data())
    assert response_data['expiry']
    response_expiry = parse_date(response_data['expiry'])
    expiry = start_datetime + timedelta(minutes=int(minutes))
    assert response_expiry <= expiry


@when('we mention user in comment for "{url}"')
def we_mention_user_in_comment(context, url):
    with context.app.mail.record_messages() as outbox:
        step_impl_when_post_url(context, url)
        assert len(outbox) == 1
        assert_equal(outbox[0].subject, "You were mentioned in a comment by test_user")
        email_text = outbox[0].body
        assert email_text


@when('we change user status to "{status}" using "{url}"')
def we_change_user_status(context, status, url):
    with context.app.mail.record_messages() as outbox:
        step_impl_when_patch_url(context, url)
        assert len(outbox) == 1
        assert_equal(outbox[0].subject, "Your Superdesk account is " + status)
        assert outbox[0].body


@when('we get the default incoming stage')
def we_get_default_incoming_stage(context):
    data = json.loads(context.response.get_data())
    incoming_stage = data['_items'][0]['incoming_stage'] if '_items' in data else data['incoming_stage']
    assert incoming_stage
    url = 'stages/{0}'.format(incoming_stage)
    when_we_get_url(context, url)
    assert_200(context.response)
    data = json.loads(context.response.get_data())
    assert data['default_incoming'] is True
    assert data['name'] == 'Incoming Stage'


@then('we get stage filled in to default_incoming')
def we_get_stage_filled_in(context):
    data = json.loads(context.response.get_data())
    assert data['task']['stage']


@given('we have sessions "{url}"')
def we_have_sessions_get_id(context, url):
    when_we_get_url(context, url)
    item = json.loads(context.response.get_data())
    context.session_id = item['_items'][0]['_id']
    context.data = item
    set_placeholder(context, 'SESSION_ID', item['_items'][0]['_id'])
    setattr(context, 'users', item['_items'][0]['user'])


@then('we get session by id')
def we_get_session_by_id(context):
    url = 'sessions/' + context.session_id
    when_we_get_url(context, url)
    item = json.loads(context.response.get_data())
    returned_id = item["_id"]
    assert context.session_id == returned_id


@then('we delete session by id')
def we_delete_session_by_id(context):
    url = 'sessions/' + context.session_id
    step_impl_when_delete_url(context, url)
    assert_200(context.response)


@when('we create a new user')
def step_create_a_user(context):
    data = apply_placeholders(context, context.text)
    with context.app.mail.record_messages() as outbox:
        context.response = context.client.post(get_prefixed_url(context.app, '/users'),
                                               data=data, headers=context.headers)
        expect_status_in(context.response, (200, 201))
        assert len(outbox) == 1
        context.email = outbox[0]


@then('we get activation email')
def step_get_activation_email(context):
    assert context.email.subject == 'Superdesk account created'
    email_text = context.email.body
    words = email_text.split()
    url = urlparse(words[words.index("to") + 1])
    token = url.fragment.split('token=')[-1]
    assert token


@then('we set elastic limit')
def step_set_limit(context):
    context.app.settings['MAX_SEARCH_DEPTH'] = 1


@then('we get emails')
def step_we_get_email(context):
    data = json.loads(context.text)
    for email in data:
        assert check_if_email_sent(context, email['body'])


@then('we get {count} emails')
def step_we_get_no_email(context, count):
    assert len(context.outbox) == int(count)


def check_if_email_sent(context, body):
    if context.outbox:
        for email in context.outbox:
            if body in email.body:
                return True
        return False


@then('we get activity')
def then_we_get_activity(context):
    url = apply_placeholders(context, '/activity?where={"name": {"$in": ["notify", "user:mention" , "desk:mention"]}}')
    context.response = context.client.get(get_prefixed_url(context.app, url), headers=context.headers)
    if context.response.status_code == 200:
        expect_json_length(context.response, 1, path='_items')
        item = json.loads(context.response.get_data())
        item = item['_items'][0]
        if item.get('_id'):
            setattr(context, 'activity', item)
            set_placeholder(context, 'USERS_ID', item['user'])


def login_as(context, username, password, user_type):
    user = {'username': username, 'password': password, 'is_active': True,
            'is_enabled': True, 'needs_activation': False, user_type: user_type}

    if context.text:
        user.update(json.loads(context.text))

    setup_auth_user(context, user)


@given('we login as user "{username}" with password "{password}" and user type "{user_type}"')
def given_we_login_as_user(context, username, password, user_type):
    login_as(context, username, password, user_type)


@when('we login as user "{username}" with password "{password}" and user type "{user_type}"')
def when_we_login_as_user(context, username, password, user_type):
    login_as(context, username, password, user_type)


def is_user_resource(resource):
    return resource in ('users', '/users')


@then('we get {no_of_stages} invisible stages')
def when_we_get_invisible_stages(context, no_of_stages):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        stages = get_resource_service('stages').get_stages_by_visibility(is_visible=False)
        assert len(stages) == int(no_of_stages)


@then('we get {no_of_stages} visible stages')
def when_we_get_visible_stages(context, no_of_stages):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        stages = get_resource_service('stages').get_stages_by_visibility(is_visible=True)
        assert len(stages) == int(no_of_stages)


@then('we get {no_of_stages} invisible stages for user')
def when_we_get_invisible_stages_for_user(context, no_of_stages):
    data = json.loads(apply_placeholders(context, context.text))
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        stages = get_resource_service('users').get_invisible_stages(data['user'])
        assert len(stages) == int(no_of_stages)


@then('we get "{field_name}" populated')
def then_field_is_populated(context, field_name):
    resp = parse_json_response(context.response)
    assert resp[field_name].get('user', None) is not None, 'item is not populated'


@then('we get "{field_name}" not populated')
def then_field_is_not_populated(context, field_name):
    resp = parse_json_response(context.response)
    assert resp[field_name] is None, 'item is not populated'


@then('we get "{field_name}" not populated in results')
def then_field_is_not_populated(context, field_name):
    resps = parse_json_response(context.response)
    for resp in resps['_items']:
        assert resp[field_name] is None, 'item is not populated'


@when('we delete content filter "{name}"')
def step_delete_content_filter(context, name):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        filter = get_resource_service('content_filters').find_one(req=None, name=name)
        url = '/content_filters/{}'.format(filter['_id'])
        headers = if_match(context, filter.get('_etag'))
        context.response = context.client.delete(get_prefixed_url(context.app, url), headers=headers)


@when('we rewrite "{item_id}"')
def step_impl_when_rewrite(context, item_id):
    context_data = {}
    _id = apply_placeholders(context, item_id)
    if context.text:
        context_data.update(json.loads(apply_placeholders(context, context.text)))
    data = json.dumps(context_data)
    context.response = context.client.post(
        get_prefixed_url(context.app, '/archive/{}/rewrite'.format(_id)),
        data=data, headers=context.headers)
    if context.response.status_code == 400:
        return

    resp = parse_json_response(context.response)
    set_placeholder(context, 'REWRITE_OF', _id)
    set_placeholder(context, 'REWRITE_ID', resp['_id']['_id'])


@when('we publish "{item_id}" with "{pub_type}" type and "{state}" state')
def step_impl_when_publish_url(context, item_id, pub_type, state):
    item_id = apply_placeholders(context, item_id)
    res = get_res('/archive/' + item_id, context)
    headers = if_match(context, res.get('_etag'))
    context_data = {"state": state}
    if context.text:
        data = apply_placeholders(context, context.text)
        context_data.update(json.loads(data))
    data = json.dumps(context_data)
    context.response = context.client.patch(get_prefixed_url(context.app, '/archive/{}/{}'.format(pub_type, item_id)),
                                            data=data, headers=headers)
    store_placeholder(context, 'archive_{}'.format(pub_type))
    resp = parse_json_response(context.response)
    linked_packages = resp.get('linked_in_packages', [])
    if linked_packages:
        take_package = linked_packages[0].get('package', '')
        set_placeholder(context, 'archive.{}.take_package'.format(item_id), take_package)


@when('we get digital item of "{item_id}"')
def step_impl_when_we_get_digital(context, item_id):
    item_id = apply_placeholders(context, item_id)
    context.response = context.client.get(get_prefixed_url(context.app, '/archive/{}'.format(item_id)),
                                          headers=context.headers)
    resp = parse_json_response(context.response)
    linked_packages = resp.get('linked_in_packages', [])
    for lp in linked_packages:
        if lp.get('package_type', '') == 'takes':
            take_package = lp.get('package', '')
            set_placeholder(context, 'archive.{}.take_package'.format(item_id), take_package)


@then('the ingest item is routed based on routing scheme and rule "{rule_name}"')
def then_ingest_item_is_routed_based_on_routing_scheme(context, rule_name):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        validate_routed_item(context, rule_name, True)


@then('the ingest item is routed and transformed based on routing scheme and rule "{rule_name}"')
def then_ingest_item_is_routed_transformed_based_on_routing_scheme(context, rule_name):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        validate_routed_item(context, rule_name, True, True)


@then('the ingest item is not routed based on routing scheme and rule "{rule_name}"')
def then_ingest_item_is_not_routed_based_on_routing_scheme(context, rule_name):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        validate_routed_item(context, rule_name, False)


def validate_routed_item(context, rule_name, is_routed, is_transformed=False):
    data = json.loads(apply_placeholders(context, context.text))

    def validate_rule(action, state):
        for destination in rule.get('actions', {}).get(action, []):
            query = {
                'and': [
                    {'term': {'ingest_id': str(data['ingest'])}},
                    {'term': {'task.desk': str(destination['desk'])}},
                    {'term': {'task.stage': str(destination['stage'])}},
                    {'term': {'state': state}}
                ]
            }
            item = get_archive_items(query) + get_published_items(query)

            if is_routed:
                assert len(item) > 0, 'No routed items found for criteria: ' + str(query)
                assert item[0]['ingest_id'] == data['ingest']
                assert item[0]['task']['desk'] == str(destination['desk'])
                assert item[0]['task']['stage'] == str(destination['stage'])
                assert item[0]['state'] == state

                if is_transformed:
                    assert item[0]['abstract'] == 'Abstract has been updated'

                assert_items_in_package(item[0], state, str(destination['desk']), str(destination['stage']))
            else:
                assert len(item) == 0

    scheme = get_resource_service('routing_schemes').find_one(_id=data['routing_scheme'], req=None)
    rule = next((rule for rule in scheme['rules'] if rule['name'].lower() == rule_name.lower()), {})
    validate_rule('fetch', 'routed')
    validate_rule('publish', 'published')


@when('we schedule the routing scheme "{scheme_id}"')
def when_we_schedule_the_routing_scheme(context, scheme_id):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        scheme_id = apply_placeholders(context, scheme_id)
        url = apply_placeholders(context, 'routing_schemes/%s' % scheme_id)
        res = get_res(url, context)
        href = get_self_href(res, context)
        headers = if_match(context, res.get('_etag'))
        rule = res.get('rules')[0]
        now = utcnow()
        from apps.rules.routing_rules import Weekdays

        rule['schedule'] = {
            'day_of_week': [
                Weekdays.dayname(now + timedelta(days=1)),
                Weekdays.dayname(now + timedelta(days=2))
            ],
            'hour_of_day_from': '1600',
            'hour_of_day_to': '2000'
        }

        if len(res.get('rules')) > 1:
            rule = res.get('rules')[1]
            rule['schedule'] = {
                'day_of_week': [Weekdays.dayname(now)]
            }

        context.response = context.client.patch(get_prefixed_url(context.app, href),
                                                data=json.dumps({'rules': res.get('rules', [])}),
                                                headers=headers)
        assert_200(context.response)


def get_archive_items(query):
    req = ParsedRequest()
    req.max_results = 100
    req.args = {'filter': json.dumps(query)}
    return list(get_resource_service('archive').get(lookup=None, req=req))


def get_published_items(query):
    req = ParsedRequest()
    req.max_results = 100
    req.args = {'filter': json.dumps(query)}
    return list(get_resource_service('published').get(lookup=None, req=req))


def assert_items_in_package(item, state, desk, stage):
    if item.get('groups'):
        terms = [{'term': {'_id': ref.get('residRef')}}
                 for ref in [ref for group in item.get('groups', [])
                             for ref in group.get('refs', []) if 'residRef' in ref]]

        query = {'or': terms}
        items = get_archive_items(query)
        assert len(items) == len(terms)
        for item in items:
            assert item.get('state') == state
            assert item.get('task', {}).get('desk') == desk
            assert item.get('task', {}).get('stage') == stage


@given('I logout')
def logout(context):
    we_have_sessions_get_id(context, '/sessions')
    step_impl_when_delete_url(context, '/auth/{}'.format(context.session_id))
    assert_200(context.response)


@then('we get "{url}" and match')
def we_get_and_match(context, url):
    url = apply_placeholders(context, url)
    response_data = get_res(url, context)
    context_data = json.loads(apply_placeholders(context, context.text))
    assert_equal(json_match(context_data, response_data), True,
                 msg=str(context_data) + '\n != \n' + str(response_data))


@then('there is no "{key}" in response')
def there_is_no_key_in_response(context, key):
    data = get_json_data(context.response)
    assert key not in data, 'key "%s" is in %s' % (key, data)


@then('there is no "{key}" in task')
def there_is_no_key_in_preferences(context, key):
    data = get_json_data(context.response)['task']
    assert key not in data, 'key "%s" is in task' % key


@then('broadcast "{key}" has value "{value}"')
def broadcast_key_has_value(context, key, value):
    data = get_json_data(context.response).get('broadcast', {})
    value = apply_placeholders(context, value)
    if value.lower() == 'none':
        assert data[key] is None, 'key "%s" is not none and has value "%s"' % (key, data[key])
    else:
        assert data[key] == value, 'key "%s" does not have valid value "%s"' % (key, data[key])


@then('there is no "{key}" in "{namespace}" preferences')
def there_is_no_key_in_preferences(context, key, namespace):
    data = get_json_data(context.response)['user_preferences']
    assert key not in data[namespace], 'key "%s" is in %s' % (key, data[namespace])


@then('we check if article has Embargo and Ed. Note of the article has embargo indication')
def step_impl_then_check_embargo(context):
    assert_200(context.response)
    try:
        response_data = json.loads(context.response.get_data())
    except Exception:
        fail_and_print_body(context.response, 'response is not valid json')

    if response_data.get('_meta') and response_data.get('_items'):
        for item in response_data.get('_items'):
            assert_embargo(context, item)
    else:
        assert_embargo(context, response_data)


def assert_embargo(context, item):
    if not item.get('embargo'):
        fail_and_print_body(context, context.response, 'Embargo not found')

    if not item.get('ednote'):
        fail_and_print_body(context, context.response, 'Embargo indication in "Ed. Note" not found')

    assert_equal((item['ednote'].find('Embargoed') > -1), True)


@when('embargo lapses for "{item_id}"')
def embargo_lapses(context, item_id):
    item_id = apply_placeholders(context, item_id)
    item = get_res("/archive/%s" % item_id, context)

    updates = {'embargo': (utcnow() - timedelta(minutes=1))}
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        get_resource_service('archive').system_update(id=item['_id'], original=item, updates=updates)


@then('we validate the published item expiry to be after publish expiry set in desk settings {publish_expiry_in_desk}')
def validate_published_item_expiry(context, publish_expiry_in_desk):
    assert_200(context.response)
    try:
        response_data = json.loads(context.response.get_data())
    except Exception:
        fail_and_print_body(context.response, 'response is not valid json')

    if response_data.get('_meta') and response_data.get('_items'):
        for item in response_data.get('_items'):
            assert_expiry(item, publish_expiry_in_desk)
    else:
        assert_expiry(response_data, publish_expiry_in_desk)


def assert_expiry(item, publish_expiry_in_desk):
    embargo = item.get('embargo')
    actual = parse_date(item.get('expiry'))
    error_message = 'Published Item Expiry validation fails'
    publish_expiry_in_desk = int(publish_expiry_in_desk)

    if embargo:
        expected = get_expiry_date(minutes=publish_expiry_in_desk,
                                   offset=datetime.strptime(embargo, '%Y-%m-%dT%H:%M:%S%z'))
        if actual != expected:
            raise WooperAssertionError("{}. Expected: {}, Actual: {}".format(error_message, expected, actual))
    else:
        expected = get_expiry_date(minutes=publish_expiry_in_desk)
        if expected < actual:
            raise WooperAssertionError("{}. Expected: {}, Actual: {}".format(error_message, expected, actual))


@when('run import legal publish queue')
def run_import_legal_publish_queue(context):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        from apps.legal_archive import ImportLegalPublishQueueCommand
        ImportLegalPublishQueueCommand().run()


@when('we expire items')
def expire_content(context):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        ids = json.loads(apply_placeholders(context, context.text))
        expiry = utcnow() - timedelta(minutes=5)
        for item_id in ids:
            original = get_resource_service('archive').find_one(req=None, _id=item_id)
            get_resource_service('archive').system_update(item_id, {'expiry': expiry}, original)
            get_resource_service('published').update_published_items(item_id, 'expiry', expiry)

        from apps.archive.commands import RemoveExpiredContent
        RemoveExpiredContent().run()


@when('the publish schedule lapses')
def run_overdue_schedule_jobs(context):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        ids = json.loads(apply_placeholders(context, context.text))
        lapse_time = utcnow() - timedelta(minutes=5)
        updates = {'publish_schedule': lapse_time, 'state': 'published'}

        for item_id in ids:
            original = get_resource_service('archive').find_one(req=None, _id=item_id)
            get_resource_service('archive').system_update(item_id, updates, original)
            get_resource_service('published').update_published_items(item_id, 'publish_schedule', lapse_time)
            get_resource_service('published').update_published_items(item_id, 'state', 'published')


@when('we get takes package "{url}" and validate')
def we_get_takes_package_and_validate(context, url):
    with context.app.test_request_context(context.app.config['URL_PREFIX']):
        url = apply_placeholders(context, url)
        response_data = get_res(url, context)

        from apps.packages.takes_package_service import TakesPackageService
        response_data = TakesPackageService().get_take_package(response_data)

        context_data = json.loads(apply_placeholders(context, context.text))
        assert_equal(json_match(context_data, response_data), True,
                     msg=str(context_data) + '\n != \n' + str(response_data))



from .common import BasePublishService, BasePublishResource, ITEM_PUBLISH
from eve.utils import config
from superdesk.metadata.item import CONTENT_TYPE, ITEM_TYPE, ITEM_STATE, CONTENT_STATE, EMBARGO
from superdesk.metadata.packages import PACKAGE_TYPE, TAKES_PACKAGE
from superdesk.publish import SUBSCRIBER_TYPES
from superdesk import get_resource_service
from superdesk.utc import utcnow
from apps.archive.common import set_sign_off, ITEM_OPERATION
import logging

logger = logging.getLogger(__name__)


class ArchivePublishResource(BasePublishResource):
    def __init__(self, endpoint_name, app, service):
        super().__init__(endpoint_name, app, service, 'publish')


class ArchivePublishService(BasePublishService):
    publish_type = 'publish'
    published_state = 'published'

    def on_update(self, updates, original):
        super().on_update(updates, original)
        set_sign_off(updates, original)

    def set_state(self, original, updates):
        """
        Set the state of the document to schedule if the publish_schedule is specified.
        :param dict original: original document
        :param dict updates: updates related to original document
        """

        updates.setdefault(ITEM_OPERATION, ITEM_PUBLISH)
        if original.get('publish_schedule') or updates.get('publish_schedule'):
            updates[ITEM_STATE] = CONTENT_STATE.SCHEDULED
        else:
            super().set_state(original, updates)

    def get_subscribers(self, doc, target_media_type):
        """
        Get the subscribers for this document based on the target_media_type for publishing.
        1. If the item has embargo and is a future date then fetch active Wire Subscribers.
           Otherwise get all active subscribers.
            a. Get the list of takes subscribers if Takes Package
        2. If targeted_for is set then exclude internet/digital subscribers.
        3. If takes package then subsequent takes are sent to same wire subscriber as first take.
        4. Filter the subscriber list based on the publish filter and global filters (if configured).
            a. Publish to takes package subscribers if the takes package is received by the subscriber.
        :param dict doc: Document to publish/correct/kill
        :param str target_media_type: dictate if the doc being queued is a Takes Package or an Individual Article.
                Valid values are - Wire, Digital. If Digital then the doc being queued is a Takes Package and if Wire
                then the doc being queues is an Individual Article.
        :return: (list, list) List of filtered subscriber,
                List of subscribers that have not received item previously (empty list in this case).
        """
        subscribers, subscribers_yet_to_receive, takes_subscribers = [], [], []
        first_take = None

        # Step 1
        query = {'is_active': True}
        if doc.get(EMBARGO) and doc.get(EMBARGO) > utcnow():
            query['subscriber_type'] = SUBSCRIBER_TYPES.WIRE

        subscribers = list(get_resource_service('subscribers').get(req=None, lookup=query))

        if doc.get(ITEM_TYPE) in [CONTENT_TYPE.COMPOSITE] and doc.get(PACKAGE_TYPE) == TAKES_PACKAGE:
            # Step 1a
            query = {'$and': [{'item_id': doc[config.ID_FIELD]},
                              {'publishing_action': {'$in': [CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED]}}]}
            takes_subscribers = self._get_subscribers_for_previously_sent_items(query)

        # Step 2
        if doc.get('targeted_for'):
            subscribers = list(self.non_digital(subscribers))

        # Step 3
        if doc.get(ITEM_TYPE) in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED]:
            first_take = self.takes_package_service.get_first_take_in_takes_package(doc)
            if first_take:
                # if first take is published then subsequent takes should to same subscribers.
                query = {'$and': [{'item_id': first_take},
                                  {'publishing_action': {'$in': [CONTENT_STATE.PUBLISHED]}}]}
                subscribers = self._get_subscribers_for_previously_sent_items(query)

        # Step 4
        if not first_take:
            subscribers = self.filter_subscribers(doc, subscribers,
                                                  SUBSCRIBER_TYPES.WIRE if doc.get('targeted_for')
                                                  else target_media_type)

        if takes_subscribers:
            # Step 4a
            subscribers_ids = set(s[config.ID_FIELD] for s in takes_subscribers)
            subscribers = takes_subscribers + [s for s in subscribers if s[config.ID_FIELD] not in subscribers_ids]

        return subscribers, subscribers_yet_to_receive



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging

from bson.objectid import ObjectId

import superdesk
from copy import deepcopy
from flask import current_app as app
from eve.utils import ParsedRequest
from eve.versioning import versioned_id_field
from superdesk.celery_app import celery
from superdesk import get_resource_service, config
from superdesk.celery_task_utils import get_lock_id
from .resource import LEGAL_ARCHIVE_NAME, LEGAL_ARCHIVE_VERSIONS_NAME, LEGAL_PUBLISH_QUEUE_NAME
from superdesk.users.services import get_display_name
from apps.archive.common import ARCHIVE
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE
from superdesk.lock import lock, unlock

logger = logging.getLogger(__name__)


class LegalArchiveImport:

    log_msg_format = "{{'_id': {_id}, 'unique_name': {unique_name}, 'version': {_current_version}, " \
                     "'expired_on': {expiry}}}."

    def upsert_into_legal_archive(self, doc):
        """
        Once publish actions are performed on the article do the below:
            1.  Get legal archive article.
            2.  De-normalize the expired article
            3.  Upserting Legal Archive.
            4.  Get Version History and De-normalize and Inserting Legal Archive Versions
        :param dict doc: doc from 'archive' collection.
        """

        if not doc.get(ITEM_STATE) in {CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED, CONTENT_STATE.KILLED}:
            logger.error('Invalid state: {}. Cannot move the item to legal archive. item: {}'.
                         format(doc.get(ITEM_STATE), self.log_msg_format.format(**doc)))
            return

        # required for behave test.
        legal_archive_doc = deepcopy(doc)

        legal_archive_service = get_resource_service(LEGAL_ARCHIVE_NAME)
        legal_archive_versions_service = get_resource_service(LEGAL_ARCHIVE_VERSIONS_NAME)

        log_msg = self.log_msg_format.format(**legal_archive_doc)
        version_id_field = versioned_id_field(app.config['DOMAIN'][ARCHIVE])
        logger.info('Preparing Article to be inserted into Legal Archive ' + log_msg)

        # Removing Irrelevant properties
        legal_archive_doc.pop(config.ETAG, None)
        legal_archive_doc.pop('lock_user', None)
        legal_archive_doc.pop('lock_session', None)
        legal_archive_doc.pop('lock_time', None)

        logger.info('Removed irrelevant properties from the article {}'.format(log_msg))

        # Step 1
        article_in_legal_archive = legal_archive_service.find_one(req=None, _id=legal_archive_doc[config.ID_FIELD])

        # Step 2 - De-normalizing the legal archive doc
        self._denormalize_user_desk(legal_archive_doc, log_msg)
        logger.info('De-normalized article {}'.format(log_msg))

        # Step 3 - Upserting Legal Archive
        logger.info('Upserting Legal Archive Repo with article {}'.format(log_msg))

        if article_in_legal_archive:
            legal_archive_service.put(legal_archive_doc[config.ID_FIELD], legal_archive_doc)
        else:
            legal_archive_service.post([legal_archive_doc])

        # Step 4 - Get Version History and De-normalize and Inserting Legal Archive Versions
        lookup = {version_id_field: legal_archive_doc[config.ID_FIELD]}
        version_history = list(get_resource_service('archive_versions').get(req=None, lookup=lookup))
        legal_version_history = list(legal_archive_versions_service.get(req=None, lookup=lookup))

        logger.info('Fetched version history for article {}'.format(log_msg))
        versions_to_insert = [version for version in version_history
                              if not any(legal_version for legal_version in legal_version_history
                                         if version[config.VERSION] == legal_version[config.VERSION])]

        # This happens when user kills an article from Dusty Archive
        if article_in_legal_archive and article_in_legal_archive[config.VERSION] < legal_archive_doc[config.VERSION] \
                and len(versions_to_insert) == 0:
            resource_def = app.config['DOMAIN'][ARCHIVE]
            versioned_doc = deepcopy(legal_archive_doc)
            versioned_doc[versioned_id_field(resource_def)] = legal_archive_doc[config.ID_FIELD]
            versioned_doc[config.ID_FIELD] = ObjectId()
            versions_to_insert.append(versioned_doc)

        for version_doc in versions_to_insert:
            self._denormalize_user_desk(version_doc,
                                        self.log_msg_format.format(_id=version_doc[version_id_field],
                                                                   unique_name=version_doc.get('unique_name'),
                                                                   _current_version=version_doc[config.VERSION],
                                                                   expiry=version_doc.get('expiry')))
            version_doc.pop(config.ETAG, None)

        if versions_to_insert:
            legal_archive_versions_service.post(versions_to_insert)
            logger.info('Inserted de-normalized version history for article {}'.format(log_msg))

        logger.info('Upsert completed for article ' + log_msg)

    def _denormalize_user_desk(self, legal_archive_doc, log_msg):
        """
        De-normalizes user, desk and stage details in legal_archive_doc.
        """

        # De-normalizing User Details
        legal_archive_doc['original_creator'] = self.__get_user_name(legal_archive_doc.get('original_creator'))
        legal_archive_doc['version_creator'] = self.__get_user_name(legal_archive_doc.get('version_creator'))

        logger.info('De-normalized User Details for article {}'.format(log_msg))

        # De-normalizing Desk and Stage details
        if legal_archive_doc.get('task'):
            if legal_archive_doc['task'].get('desk'):
                desk = get_resource_service('desks').find_one(req=None, _id=str(legal_archive_doc['task']['desk']))
                if desk:
                    legal_archive_doc['task']['desk'] = desk.get('name')
                    logger.info('De-normalized Desk Details for article {}'.format(log_msg))
                else:
                    logger.info('Desk Details Not Found: {}. {}'.format(legal_archive_doc['task'].get('desk'), log_msg))

            if legal_archive_doc['task'].get('stage'):
                stage = get_resource_service('stages').find_one(req=None, _id=str(legal_archive_doc['task']['stage']))
                if stage:
                    legal_archive_doc['task']['stage'] = stage.get('name')
                    logger.info('De-normalized Stage Details for article {}'.format(log_msg))
                else:
                    logger.info('Stage Details Not Found: {}. {}'.format(legal_archive_doc['task'].get('stage'),
                                                                         log_msg))

            legal_archive_doc['task']['user'] = self.__get_user_name(legal_archive_doc['task'].get('user'))

    def __get_user_name(self, user_id):
        """
        Retrieves display_name of the user identified by user_id
        """
        logger.info('Get User Details for ID:{}'.format(user_id))

        if not user_id:
            return ''

        user = get_resource_service('users').find_one(req=None, _id=user_id)

        if not user:
            return ''

        return get_display_name(user)

    def import_legal_publish_queue(self):
        """
        Import legal publish queue.
        """
        logger.info('Starting to import publish queue items...')
        max_date = self._get_max_date_from_publish_queue()
        logger.info('Get publish queue items from datetime - {}'.format(max_date))
        queue_items = list(self._get_publish_queue_items_to_import(max_date))

        if len(queue_items) == 0:
            logger.info('No Items to import.')
            return

        logger.info('Items to import {}.'.format(len(queue_items)))
        logger.info('Get subscribers info for de-normalising queue items.')
        subscriber_ids = list({str(queue_item['subscriber_id']) for queue_item in queue_items})
        query = {'$and': [{config.ID_FIELD: {'$in': subscriber_ids}}]}
        subscribers = list(get_resource_service('subscribers').get(req=None, lookup=query))
        subscribers = {str(subscriber[config.ID_FIELD]): subscriber for subscriber in subscribers}

        for queue_item in queue_items:
            try:
                self._upsert_into_legal_archive_publish_queue(queue_item, subscribers)
            except:
                logger.exception("Failed to import publish queue item. {}".format(queue_item.get(config.ID_FIELD)))

        logger.info('Completed importing of publish queue items {}.'.format(max_date))

    def _get_publish_queue_items_to_import(self, max_date):
        """
        Get the queue items to import after max_date
        :param datetime max_date:
        :return : list of publish queue items
        """
        legal_publish_queue_service = get_resource_service('publish_queue')

        lookup = {}
        if max_date:
            lookup[config.LAST_UPDATED] = {'$gte': max_date}

        req = ParsedRequest()
        req.max_results = 500

        return legal_publish_queue_service.get(req=req, lookup=lookup)

    def _get_max_date_from_publish_queue(self):
        """
        Get the max _updated date from legal_publish_queue collection
        :return datetime: _updated time
        """
        legal_publish_queue_service = get_resource_service(LEGAL_PUBLISH_QUEUE_NAME)
        req = ParsedRequest()
        req.sort = '[("%s", -1)]' % config.LAST_UPDATED
        req.max_results = 1
        queue_item = list(legal_publish_queue_service.get(req=req, lookup={}))
        return queue_item[0][config.LAST_UPDATED] if queue_item else None

    def _upsert_into_legal_archive_publish_queue(self, queue_item, subscribers):
        """
        Upsert into legal publish queue.
        :param dict queue_item: publish_queue collection item
        :param dict subscribers: subscribers information
        """
        legal_publish_queue_service = get_resource_service(LEGAL_PUBLISH_QUEUE_NAME)
        legal_queue_item = queue_item.copy()
        lookup = {
            'item_id': legal_queue_item['item_id'],
            'item_version': legal_queue_item['item_version'],
            'subscriber_id': legal_queue_item['subscriber_id']
        }

        log_msg = '{item_id} -- version {item_version} -- subscriber {subscriber_id}.'.format(**lookup)

        logger.info('Processing queue item: {}'.format(log_msg))

        existing_queue_item = legal_publish_queue_service.find_one(req=None, _id=legal_queue_item.get(config.ID_FIELD))
        legal_queue_item['subscriber_id'] = subscribers[str(queue_item['subscriber_id'])]['name']
        legal_queue_item['_subscriber_id'] = queue_item['subscriber_id']

        legal_queue_item.pop(config.ETAG, None)

        if not existing_queue_item:
            legal_publish_queue_service.post([legal_queue_item])
            logger.info('Inserted queue item: {}'.format(log_msg))
        else:
            legal_publish_queue_service.put(existing_queue_item.get(config.ID_FIELD), legal_queue_item)
            logger.info('Updated queue item: {}'.format(log_msg))

        logger.info('Processed queue item: {}'.format(log_msg))


@celery.task(bind=True)
def import_into_legal_archive(self, doc):
    """
    Called async to import into legal archive.
    :param self: celery task
    :param dict doc: document to import into legal_archive
    """
    try:
        LegalArchiveImport().upsert_into_legal_archive(doc)
    except:
        logger.exception("Failed to import into legal archive.")


class ImportLegalPublishQueueCommand(superdesk.Command):
    """
    This command import publish queue records into legal publish queue.
    """
    def run(self):
        logger.info('Importing Legal Publish Queue')
        lock_name = get_lock_id('legal_archive', 'import_legal_publish_queue')
        if not lock(lock_name, '', expire=600):
            return
        try:
            LegalArchiveImport().import_legal_publish_queue()
        finally:
            unlock(lock_name, '')


superdesk.command('legal_publish_queue:import', ImportLegalPublishQueueCommand())



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import superdesk

from superdesk import get_resource_service, config
from superdesk.utc import utcnow
from superdesk.errors import SubscriberError, SuperdeskPublishError, PublishQueueError


logger = logging.getLogger(__name__)


class PublishService():
    """Base publish service class."""

    def _transmit(self, queue_item, subscriber):
        """Performs the publishing of the queued item. Implement in subclass

        @param queue_item: the queued item document
        @type queue_item: dict
        @param subscriber: the subscriber document
        @type subscriber: dict
        """
        raise NotImplementedError()

    def transmit(self, queue_item):
        subscriber = get_resource_service('subscribers').find_one(req=None, _id=queue_item['subscriber_id'])

        if not subscriber.get('is_active'):
            raise SubscriberError.subscriber_inactive_error(Exception('Subscriber inactive'), subscriber)
        else:
            try:
                self._transmit(queue_item, subscriber) or []
                update_item_status(queue_item, 'success')
            except SuperdeskPublishError as error:
                update_item_status(queue_item, 'error', error)
                self.close_transmitter(subscriber, error)
                raise error

    def close_transmitter(self, subscriber, error):
        """
        Checks if the transmitter has the error code set in the list of critical errors then closes the transmitter.

        :param error: The error thrown during transmission
        """

        if subscriber.get('critical_errors', {}).get(str(error.code)):
            update = {
                'is_active': False,
                'last_closed': {
                    'closed_at': utcnow(),
                    'message': 'Subscriber made inactive due to critical error: {}'.format(error)
                }
            }

            get_resource_service('subscribers').system_update(subscriber[config.ID_FIELD], update, subscriber)


def update_item_status(queue_item, status, error=None):
    try:
        item_update = {'state': status}
        if status == 'in-progress':
            item_update['transmit_started_at'] = utcnow()
        elif status == 'success':
            item_update['completed_at'] = utcnow()
        elif status == 'error' and error:
            item_update['error_message'] = '{}:{}'.format(error, str(error.system_exception))

        superdesk.get_resource_service('publish_queue').patch(queue_item.get('_id'), item_update)
    except Exception as ex:
        raise PublishQueueError.item_update_error(ex)


def get_file_extension(queue_item):
    try:
        format = queue_item['destination']['format'].upper()
        if format == 'NITF':
            return 'ntf'
        if format == 'XML':
            return 'xml'
    except Exception as ex:
        raise PublishQueueError.item_update_error(ex)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

''' Amazon media storage module'''
import boto3
import json
import bson
import time
import logging
from io import BytesIO
from eve.io.media import MediaStorage
from superdesk.media.media_operations import download_file_from_url
from superdesk.upload import upload_url

logger = logging.getLogger(__name__)
MAX_KEYS = 1000


class AmazonObjectWrapper(BytesIO):

    def __init__(self, s3_object, name, metadata):
        super().__init__()

        s3_body = s3_object['Body']
        blocksize = 65636
        buf = s3_body.read(amt=blocksize)
        while len(buf) > 0:
            self.write(buf)
            buf = s3_body.read(amt=blocksize)

        self.seek(0)
        self.content_type = s3_object['ContentType']
        self.length = int(s3_object['ContentLength'])
        self.name = name
        self.filename = name
        self.metadata = metadata
        self.upload_date = s3_object['LastModified']
        self.md5 = s3_object['ETag'][1:-1]


class AmazonMediaStorage(MediaStorage):

    def __init__(self, app=None):
        super().__init__(app)
        username, api_key = self.read_from_config()
        self.client = boto3.client('s3',
                                   aws_access_key_id=username,
                                   aws_secret_access_key=api_key,
                                   region_name=self.region)
        self.user_metadata_header = 'x-amz-meta-'

    def url_for_media(self, media_id):
        if not self.app.config.get('AMAZON_SERVE_DIRECT_LINKS', False):
            return upload_url(str(media_id))
        protocol = 'https' if self.app.config.get('AMAZON_S3_USE_HTTPS', False) else 'http'
        endpoint = 's3-%s.amazonaws.com' % self.app.config.get('AMAZON_REGION')
        return '%s://%s.%s/%s' % (protocol, self.container_name, endpoint, media_id)

    def media_id(self, filename):
        if not self.app.config.get('AMAZON_SERVE_DIRECT_LINKS', False):
            return str(bson.ObjectId())
        return '%s/%s' % (time.strftime('%Y%m%d'), filename)

    def fetch_rendition(self, rendition):
        stream, name, mime = download_file_from_url(rendition.get('href'))
        return stream

    def read_from_config(self):
        self.region = self.app.config.get('AMAZON_REGION', 'us-east-1') or 'us-east-1'
        username = self.app.config['AMAZON_ACCESS_KEY_ID']
        api_key = self.app.config['AMAZON_SECRET_ACCESS_KEY']
        self.container_name = self.app.config['AMAZON_CONTAINER_NAME']
        self.kwargs = {}
        if self.app.config.get('AMAZON_SERVE_DIRECT_LINKS', False):
            self.kwargs['ACL'] = 'public-read'
        return username, api_key

    def get(self, id_or_filename, resource=None):
        """ Opens the file given by name or unique id. Note that although the
        returned file is guaranteed to be a File object, it might actually be
        some subclass. Returns None if no file was found.
        """
        id_or_filename = str(id_or_filename)
        try:
            obj = self.client.get_object(Key=id_or_filename, Bucket=self.container_name)
            if obj:
                metadata = self.extract_metadata_from_headers(obj['Metadata'])
                return AmazonObjectWrapper(obj, id_or_filename, metadata)
        except:
            return None
        return None

    def get_all_keys(self):
        """ Returns the list of all keys from the bucket """
        all_keys = []
        try:
            for objects in self._get_all_keys_in_batches():
                all_keys.extend(objects)
        except Exception as ex:
            logger.exception(ex)
        finally:
            return all_keys

    def _get_all_keys_in_batches(self):
        """ Returns the list of all keys from the bucket in batches """
        NextMarker = ''
        while True:
            objects = self.client.list_objects(Bucket=self.container_name, Marker=NextMarker, MaxKeys=MAX_KEYS)

            if not objects or len(objects.get('Contents', [])) == 0:
                return

            keys = [obj['Key'] for obj in objects.get('Contents', [])]
            NextMarker = keys[-1]
            yield keys

    def extract_metadata_from_headers(self, request_headers):
        headers = {}
        for key, value in request_headers.items():
            if self.user_metadata_header in key:
                new_key = key.split(self.user_metadata_header)[1]
                if(value):
                    try:
                        headers[new_key] = json.loads(value)
                    except Exception as ex:
                        logger.exception(ex)
        return headers

    def transform_metadata_to_amazon_format(self, metadata):
        if not metadata:
            return {}
        file_metadata = {}
        for key, value in metadata.items():
            new_key = self.user_metadata_header + key
            file_metadata[new_key] = value
        return file_metadata

    def put(self, content, filename=None, content_type=None, resource=None, metadata=None, _id=None):
        """ Saves a new file using the storage system, preferably with the name
        specified. If there already exists a file with this name name, the
        storage system may modify the filename as necessary to get a unique
        name. Depending on the storage system, a unique id or the actual name
        of the stored file will be returned. The content type argument is used
        to appropriately identify the file when it is retrieved.
        """
        logger.debug('Going to save file file=%s media=%s ' % (filename, _id))
        _id = _id or self.media_id(filename)
        found = self._check_exists(_id)
        if found:
            return _id

        try:
            file_metadata = self.transform_metadata_to_amazon_format(metadata)
            self.client.put_object(Key=_id, Body=content, Bucket=self.container_name,
                                   ContentType=content_type, Metadata=file_metadata, **self.kwargs)
            return _id
        except Exception as ex:
            logger.exception(ex)
            raise

    def delete(self, id_or_filename, resource=None):
        id_or_filename = str(id_or_filename)
        del_res = self.client.delete_object(Key=id_or_filename, Bucket=self.container_name)
        logger.debug('Amazon S3 file deleted %s with status' % id_or_filename, del_res)

    def delete_objects(self, ids):
        """ Deletes the objects with given list of ids"""
        try:
            delete_parameters = {'Objects': [{'Key': id} for id in ids], 'Quiet': True}
            response = self.client.delete_objects(Bucket=self.container_name, Delete=delete_parameters)
            if len(response.get('Errors', [])):
                errors = ','.join(['{}:{}'.format(error['Key'], error['Message']) for error in response['Errors']])
                logger.error('Files couldn\'t be deleted: {}'.format(errors))
                return False, errors
            return True, None
        except Exception as ex:
            logger.exception(ex)
            raise

    def exists(self, id_or_filename, resource=None):
        """ Returns True if a file referenced by the given name or unique id
        already exists in the storage system, or False if the name is available
        for a new file.
        """
        id_or_filename = str(id_or_filename)
        found = self._check_exists(id_or_filename)
        return found

    def _check_exists(self, id_or_filename):
        try:
            self.client.head_object(Key=id_or_filename, Bucket=self.container_name)
            return True
        except Exception:
            # File not found
            return False

    def remove_unreferenced_files(self, existing_files):
        """ Gets the files from S3 and compares against existing and deletes the orphans """
        bucket_files = self.get_all_keys()
        orphan_files = list(set(bucket_files) - existing_files)
        print('There are {} orphan files...'.format(len(orphan_files)))

        if len(orphan_files) > 0:
            print('Cleaning the orphan files...')
            deleted, errors = self.delete_objects(orphan_files)
            if deleted:
                print('Image cleaning completed successfully.')
            else:
                print('Failed to clean orphans: {}'.format(errors))
        else:
            print('There\'s nothing to clean.')



from eve.io.base import DataLayer
from io import BytesIO
import urllib3
import json
import datetime
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE
from superdesk.utc import utc, utcnow
from eve_elastic.elastic import ElasticCursor
from superdesk.media.media_operations import process_file_from_stream, decode_metadata
from superdesk.media.renditions import generate_renditions, delete_file_on_error
from superdesk.errors import SuperdeskApiError
from superdesk.io.iptc import subject_codes
from flask import url_for
import urllib

import logging


urllib3.disable_warnings()

logger = logging.getLogger(__name__)


class AAPMMDatalayer(DataLayer):
    def __set_auth_cookie(self, app):
        if self._username is not None and self._password is not None:
            url = app.config['AAP_MM_SEARCH_URL'] + '/Users/login'
            values = {'username': self._username, 'password': self._password}
            r = self._http.urlopen('POST', url, headers={'Content-Type': 'application/json'}, body=json.dumps(values))
        else:
            url = app.config['AAP_MM_SEARCH_URL'] + '/Users/AnonymousToken'
            r = self._http.request('GET', url, redirect=False)

        self._headers = {'cookie': r.getheader('set-cookie'), 'Content-Type': 'application/json'}

    def set_credentials(self, user, password):
        if user != self._username and user != '' and password != self._password and password != '':
            self._username = user
            self._password = password
            self.__set_auth_cookie(self._app)

    def init_app(self, app):
        app.config.setdefault('AAP_MM_SEARCH_URL', 'https://one-api.aap.com.au/api/v3')
        app.config.setdefault('AAP_MM_CDN_URL', 'http://one-cdn.aap.com.au/Preview.mp4')
        self._app = app
        self._headers = None
        self._username = None
        self._password = None
        self._http = urllib3.PoolManager()

    def find(self, resource, req, lookup):
        """
        Called to execute a search against the AAP Mulitmedia API. It attempts to translate the search request
        passed in req to a suitable form for a search request against the API. It parses the response into a
        suitable ElasticCursor, the front end will never know.
        :param resource:
        :param req:
        :param lookup:
        :return:
        """
        if self._headers is None:
            self.__set_auth_cookie(self._app)

        url = self._app.config['AAP_MM_SEARCH_URL'] + '/Assets/search'
        query_keywords = '*:*'
        if 'query' in req['query']['filtered']:
            query_keywords = req['query']['filtered']['query']['query_string']['query']
            query_keywords = query_keywords.replace('slugline:', 'objectname:')
            query_keywords = query_keywords.replace('description:', 'captionabstract:')

        fields = {}
        for criterion in req.get('post_filter', {}).get('and', {}):
            # parse out the date range if possible
            if 'range' in criterion:
                start = None
                end = None
                daterange = None
                if 'firstcreated' in criterion.get('range', {}):
                    if 'gte' in criterion['range']['firstcreated']:
                        start = criterion['range']['firstcreated']['gte'][0:10]
                    if 'lte' in criterion['range']['firstcreated']:
                        end = criterion['range']['firstcreated']['lte'][0:10]
                # if there is a special start and no end it's one of the date buttons
                if start and not end:
                    if start == 'now-24H':
                        daterange = {'Dates': ['[NOW/HOUR-24HOURS TO NOW/HOUR]']}
                    if start == 'now-1w':
                        daterange = {'Dates': ['[NOW/DAY-7DAYS TO NOW/DAY]']}
                    if start == 'now-1M':
                        daterange = {'Dates': ['[NOW/DAY-1MONTH TO NOW/DAY]']}
                # we've got something but no daterange set above
                if (start or end) and not daterange:
                    daterange = {'DateRange': [{'Start': start, 'End': end}], 'DateCreatedFilter': 'true'}
                if daterange:
                    fields.update(daterange)

            if 'terms' in criterion:
                if 'type' in criterion.get('terms', {}):
                    fields.update({'MediaTypes': criterion['terms']['type']})
                if 'credit' in criterion.get('terms', {}):
                    fields.update({'Credits': criterion['terms']['credit']})
                if 'anpa_category.name' in criterion.get('terms', {}):
                    cat_list = []
                    for cat in criterion['terms']['anpa_category.name']:
                        qcode = [key for key, value in subject_codes.items() if value == cat]
                        if qcode:
                            for code in qcode:
                                cat_list.append(code)
                        else:
                            cat_list.append(cat)
                    fields.update({'Categories': cat_list})

        size = int(req.get('size', '25')) if int(req.get('size', '25')) > 0 else 25
        query = {'Query': query_keywords, 'pageSize': str(size),
                 'pageNumber': str(int(req.get('from', '0')) // size + 1)}

        r = self._http.urlopen('POST', url + '?' + urllib.parse.urlencode(query),
                               body=json.dumps(fields), headers=self._headers)
        hits = self._parse_hits(json.loads(r.data.decode('UTF-8')))
        return ElasticCursor(docs=hits['docs'], hits={'hits': hits, 'aggregations': self._parse_aggregations(hits)})

    def _parse_doc(self, doc):
        new_doc = {}
        new_doc['_id'] = doc['AssetId']
        new_doc['guid'] = doc['AssetId']
        new_doc['headline'] = doc['Title']
        new_doc['description'] = doc['Description']
        new_doc['source'] = doc['Credit']
        if 'Source' in doc:
            new_doc['original_source'] = doc['Credit'] + '/' + str(doc.get('Source', ''))
        else:
            new_doc['original_source'] = doc['Credit']
        new_doc['versioncreated'] = self._datetime(doc['ModifiedDate'])
        new_doc['firstcreated'] = self._datetime(doc['CreationDate'])
        new_doc['pubstatus'] = 'usable'
        # This must match the action
        new_doc['_type'] = 'externalsource'
        # entry that the client can use to identify the fetch endpoint
        new_doc['fetch_endpoint'] = 'aapmm'
        if doc['AssetType'] == 'VIDEO':
            new_doc[ITEM_TYPE] = CONTENT_TYPE.VIDEO
            purl = '{}?assetType=VIDEO&'.format(self._app.config['AAP_MM_CDN_URL'])
            purl += 'path=/rest/aap/archives/imagearc/dossiers/{}'.format(doc['AssetId'])
            purl += '/files/ipod&assetId={}&mimeType=video/mp4&dummy.mp4'.format(doc['AssetId'])
            new_doc['renditions'] = {'original': {'href': purl, 'mimetype': 'video/mp4'}}
        else:
            new_doc[ITEM_TYPE] = CONTENT_TYPE.PICTURE
            new_doc['renditions'] = {
                'viewImage': {'href': doc.get('Preview', doc.get('Layout'))['Href']},
                'thumbnail': {'href': doc.get('Thumbnail', doc.get('Layout'))['Href']},
                'original': {'href': doc.get('Preview', doc.get('Layout'))['Href']},
                'baseImage': {'href': doc.get('Preview', doc.get('Layout'))['Href']},
            }

        new_doc['slugline'] = doc['Title']
        new_doc['byline'] = doc['Byline']
        new_doc['ednote'] = doc['SpecialInstructions']
        doc.clear()
        doc.update(new_doc)

    def _parse_hits(self, hits):
        hits['docs'] = hits.pop('Assets')
        hits['total'] = hits.pop('Total')
        for doc in hits['docs']:
            self._parse_doc(doc)
        return hits

    def _parse_aggregation(self, aggregations, facet, aggregation, hits):
        """
        Converts the "facet" to the "aggregate" based on the FacetResults in hits returns the equivalent
        aggregation in aggregations
        :param aggregations:
        :param facet:
        :param aggregation:
        :param hits:
        :return:
        """
        if 'FacetResults' in hits and facet in hits.get('FacetResults', {}):
            buckets = []
            name_set = set()
            for cat in hits.get('FacetResults', {}).get(facet, {}):
                if cat['DisplayName'] in name_set:
                    buckets.append({'doc_count': cat['Count'], 'key': cat['DisplayName'] + '/' + cat['Name'],
                                    'qcode': cat['Name']})
                else:
                    buckets.append({'doc_count': cat['Count'], 'key': cat['DisplayName'], 'qcode': cat['Name']})
                    name_set.add(cat['DisplayName'])
            aggregations[aggregation] = {'buckets': buckets}

    def _parse_aggregation_dates(self, aggregations, hits):
        """
        Extract the date related facets and convert into aggregations
        :param aggregations:
        :param hits:
        :return:
        """
        if 'FacetResults' in hits and 'Dates' in hits.get('FacetResults', {}):
            for date in hits.get('FacetResults', {}).get('Dates', {}):
                if date['Name'] == '[NOW/HOUR-24HOURS TO NOW/HOUR]':
                    aggregations['day'] = {'buckets': [{'doc_count': date['Count'], 'key': date['Name']}]}
                if date['Name'] == '[NOW/DAY-7DAYS TO NOW/DAY]':
                    aggregations['week'] = {'buckets': [{'doc_count': date['Count'], 'key': date['Name']}]}
                if date['Name'] == '[NOW/DAY-1MONTH TO NOW/DAY]':
                    aggregations['month'] = {'buckets': [{'doc_count': date['Count'], 'key': date['Name']}]}

    def _parse_aggregations(self, hits):
        """
        Given the hits returned from the AAP Mulitmedia API it will convert them to the same format as the
        Aggregations returned from the superdesk search against Elastic
        :param hits:
        :return: The converted aggregations
        """
        aggregations = {}
        self._parse_aggregation(aggregations, 'Categories', 'category', hits)
        self._parse_aggregation(aggregations, 'MediaTypes', 'type', hits)
        self._parse_aggregation(aggregations, 'Credits', 'credit', hits)
        self._parse_aggregation_dates(aggregations, hits)
        hits.pop('FacetResults')
        return aggregations

    def _datetime(self, string):
        try:
            dt = datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%S').replace(tzinfo=utc)
        except:
            dt = utcnow()
        return dt

    def _get_resolutions(self, id):
        url = self._app.config['AAP_MM_SEARCH_URL'] + '/Assets/Resolutions'
        values = [id]
        headers = dict(self._headers)
        headers['Content-Type'] = 'application/json'
        r = self._http.urlopen('POST', url, headers=headers, body=json.dumps(values))
        return json.loads(r.data.decode('UTF-8'))

    def find_all(self, resource, max_results=1000):
        raise NotImplementedError

    def find_one(self, resource, req, **lookup):
        raise NotImplementedError

    def find_one_raw(self, resource, _id):
        if self._headers is None:
            self.__set_auth_cookie(self._app)

        url = self._app.config['AAP_MM_SEARCH_URL'] + '/Assets/{}'.format(_id)
        r = self._http.request('GET', url, headers=self._headers)
        doc = json.loads(r.data.decode('UTF-8'))
        self._parse_doc(doc)
        if 'fetch_endpoint' in doc:
            del doc['fetch_endpoint']

        # Only if we have credentials can we download the original if the account has that privilege
        if self._username is not None and self._password is not None:
            resolutions = self._get_resolutions(_id)
            if doc[ITEM_TYPE] == CONTENT_TYPE.PICTURE:
                if any(i['Name'] == 'Original' for i in resolutions['Image']):
                    url = self._app.config['AAP_MM_SEARCH_URL'] + '/Assets/{}/Original/download'.format(_id)
                    mime_type = 'image/jpeg'
                else:
                    raise FileNotFoundError
            elif doc[ITEM_TYPE] == CONTENT_TYPE.VIDEO:
                if any(v['Name'] == 'Ipod' for v in resolutions['Video']):
                    url = self._app.config['AAP_MM_SEARCH_URL'] + '/Assets/{}/Ipod/download'.format(_id)
                    mime_type = doc.get('renditions').get('original').get('mimetype')
                else:
                    raise FileNotFoundError
            else:
                raise NotImplementedError
        else:
            if doc[ITEM_TYPE] == CONTENT_TYPE.VIDEO:
                mime_type = doc.get('renditions').get('original').get('mimetype')
            else:
                mime_type = 'image/jpeg'
            url = doc['renditions']['original']['href']

        r = self._http.request('GET', url, headers=self._headers)
        out = BytesIO(r.data)
        file_name, content_type, metadata = process_file_from_stream(out, mime_type)

        inserted = []

        try:
            logger.debug('Going to save media file with %s ' % file_name)
            out.seek(0)
            file_id = self._app.media.put(out, filename=file_name, content_type=content_type, metadata=None)
            doc['mimetype'] = content_type
            doc['filemeta'] = decode_metadata(metadata)
            # set the version created to now to bring it to the top of the desk, images can be quite old
            doc['versioncreated'] = utcnow()
            inserted = [file_id]
            file_type = content_type.split('/')[0]
            rendition_spec = self._app.config['RENDITIONS']['picture']

            renditions = generate_renditions(out, file_id, inserted, file_type,
                                             content_type, rendition_spec,
                                             self.url_for_media, insert_metadata=False)
            doc['renditions'] = renditions
        except Exception as io:
            logger.exception(io)
            for file_id in inserted:
                delete_file_on_error(doc, file_id)

            raise SuperdeskApiError.internalError('Generating renditions failed')

        return doc

    def url_for_media(self, media_id):
        return url_for('upload_raw.get_upload_as_data_uri', media_id=media_id,
                       _external=True, _schema=self._app.config['URL_PROTOCOL'])

    def find_list_of_ids(self, resource, ids, client_projection=None):
        raise NotImplementedError

    def insert(self, resource, docs, **kwargs):
        raise NotImplementedError

    def update(self, resource, id_, updates, original):
        raise NotImplementedError

    def update_all(self, resource, query, updates):
        raise NotImplementedError

    def replace(self, resource, id_, document, original):
        raise NotImplementedError

    def remove(self, resource, lookup=None):
        raise NotImplementedError

    def is_empty(self, resource):
        raise NotImplementedError



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from unittest import TestCase
from .dpa_derive_dateline import dpa_derive_dateline


class DPADeriveDatelineTests(TestCase):

    def simple_case_test(self):
        item = dict()
        item['body_html'] = 'Madrid (dpa) - Results and standings Spanish Primera Division\r\n(kick-off times in GMT):'
        dpa_derive_dateline(item)
        self.assertEqual(item['dateline']['located']['city'], 'Madrid')

    def text_before_test(self):
        item = dict()
        item['body_html'] = '\r\nComedian Morales running as outsider in Guatemala =\r\n        \r\n'
        item['body_html'] += 'Guatemala City (dpa) - Comedian Jimmy Morales has one major asset'
        dpa_derive_dateline(item)
        self.assertEqual(item['dateline']['located']['city'], 'Guatemala City')

    def know_city_test(self):
        item = dict()
        item['body_html'] = 'Dubbo (dpa) - Where on earth is that'
        dpa_derive_dateline(item)
        self.assertEqual(item['dateline']['located']['city'], 'Dubbo')
        self.assertEqual(item['dateline']['located']['country'], 'Australia')

    def no_match_test(self):
        item = dict()
        item['body_html'] = 'No dateline here'
        dpa_derive_dateline(item)
        self.assertNotIn('dateline', item)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import textwrap
from io import StringIO

from bs4 import BeautifulSoup

from superdesk.publish.formatters import Formatter
from apps.publish.formatters.aap_formatter_common import map_priority, set_subject

import superdesk
from superdesk.errors import FormatterError
from superdesk.io.iptc import subject_codes
from apps.publish.formatters.field_mappers.selectorcode_mapper import SelectorcodeMapper
from apps.publish.formatters.field_mappers.locator_mapper import LocatorMapper
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE, EMBARGO


class AAPIpNewsFormatter(Formatter):

    def format(self, article, subscriber):
        """
        Constructs a dictionary that represents the parameters passed to the IPNews InsertNews stored procedure
        :return: returns the sequence number of the subscriber and the constructed parameter dictionary
        """
        try:
            docs = []
            for category in article.get('anpa_category'):
                pub_seq_num = superdesk.get_resource_service('subscribers').generate_sequence_number(subscriber)
                odbc_item = {'originator': article.get('source', None), 'sequence': pub_seq_num,
                             'category': category.get('qcode'),
                             'headline': article.get('headline', '').replace('\'', '\'\''),
                             'author': article.get('byline', '').replace('\'', '\'\''),
                             'keyword': self.append_legal(article=article, truncate=True).replace('\'', '\'\''),
                             'subject_reference': set_subject(category, article)}

                if 'subject_reference' in odbc_item and odbc_item['subject_reference'] is not None \
                        and odbc_item['subject_reference'] != '00000000':
                    odbc_item['subject'] = subject_codes[odbc_item['subject_reference'][:2] + '000000']
                    if odbc_item['subject_reference'][2:5] != '000':
                        odbc_item['subject_matter'] = subject_codes[odbc_item['subject_reference'][:5] + '000']
                    else:
                        odbc_item['subject_matter'] = ''
                    if not odbc_item['subject_reference'].endswith('000'):
                        odbc_item['subject_detail'] = subject_codes[odbc_item['subject_reference']]
                    else:
                        odbc_item['subject_detail'] = ''
                else:
                    odbc_item['subject_reference'] = '00000000'

                odbc_item['take_key'] = article.get('anpa_take_key', None)  # @take_key
                odbc_item['usn'] = article.get('unique_id', None)  # @usn
                if article[ITEM_TYPE] == CONTENT_TYPE.PREFORMATTED:  # @article_text
                    odbc_item['article_text'] = self.append_body_footer(article).replace('\'', '\'\'')
                elif article[ITEM_TYPE] == CONTENT_TYPE.TEXT:
                    soup = BeautifulSoup(self.append_body_footer(article), "html.parser")
                    text = StringIO()
                    for p in soup.findAll('p'):
                        text.write('\x19\r\n')
                        ptext = p.get_text('\n')
                        for l in ptext.split('\n'):
                            if len(l) > 80:
                                text.write(textwrap.fill(l, 80).replace('\n', ' \r\n'))
                            else:
                                text.write(l + ' \r\n')
                    odbc_item['article_text'] = text.getvalue().replace('\'', '\'\'')

                if 'genre' in article:
                    odbc_item['genre'] = article['genre'][0].get('name', None)
                else:
                    odbc_item['genre'] = 'Current'  # @genre
                if article.get(ITEM_TYPE, CONTENT_TYPE.TEXT) == CONTENT_TYPE.TEXT:
                    odbc_item['texttab'] = 'x'
                elif article.get(ITEM_TYPE, None) == CONTENT_TYPE.PREFORMATTED:
                    odbc_item['texttab'] = 't'
                odbc_item['wordcount'] = article.get('word_count', None)  # @wordcount
                odbc_item['news_item_type'] = 'News'
                odbc_item['priority'] = map_priority(article.get('priority'))  # @priority
                odbc_item['service_level'] = 'a'  # @service_level
                odbc_item['fullStory'] = 1
                odbc_item['ident'] = '0'  # @ident

                SelectorcodeMapper().map(article, category.get('qcode').upper(),
                                         subscriber=subscriber,
                                         formatted_item=odbc_item)
                headline_prefix = LocatorMapper().map(article, category.get('qcode').upper())
                if headline_prefix:
                    odbc_item['headline'] = '{}:{}'.format(headline_prefix, odbc_item['headline'])

                if article.get(EMBARGO):
                    embargo = '{}{}'.format('Embargo Content. Timestamp: ', article.get(EMBARGO).isoformat())
                    odbc_item['article_text'] = embargo + odbc_item['article_text']

                docs.append((pub_seq_num, odbc_item))

            return docs
        except Exception as ex:
            raise FormatterError.AAPIpNewsFormatterError(ex, subscriber)

    def can_format(self, format_type, article):
        return format_type == 'AAP IPNEWS' and article[ITEM_TYPE] in [CONTENT_TYPE.TEXT, CONTENT_TYPE.PREFORMATTED]



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
from superdesk.celery_app import celery
from superdesk import get_backend, privilege
from .resource import LegalArchiveResource, LegalArchiveVersionsResource, LegalPublishQueueResource, \
    LEGAL_ARCHIVE_NAME, LEGAL_ARCHIVE_VERSIONS_NAME, LEGAL_PUBLISH_QUEUE_NAME
from .service import LegalArchiveService, LegalArchiveVersionsService, LegalPublishQueueService
from .commands import ImportLegalPublishQueueCommand  # noqa

logger = logging.getLogger(__name__)


def init_app(app):
    endpoint_name = LEGAL_ARCHIVE_NAME
    service = LegalArchiveService(endpoint_name, backend=get_backend())
    LegalArchiveResource(endpoint_name, app=app, service=service)

    endpoint_name = LEGAL_ARCHIVE_VERSIONS_NAME
    service = LegalArchiveVersionsService(endpoint_name, backend=get_backend())
    LegalArchiveVersionsResource(endpoint_name, app=app, service=service)

    endpoint_name = LEGAL_PUBLISH_QUEUE_NAME
    service = LegalPublishQueueService(endpoint_name, backend=get_backend())
    LegalPublishQueueResource(endpoint_name, app=app, service=service)

    privilege(name=LEGAL_ARCHIVE_NAME, label='Legal Archive', description='Read from legal archive')


@celery.task
def import_legal_publish_queue():
    ImportLegalPublishQueueCommand().run()



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.publish.subscribers import SUBSCRIBER_TYPES

from test_factory import SuperdeskTestCase
from apps.publish import init_app
from apps.publish.formatters.anpa_formatter import AAPAnpaFormatter
from apps.publish.formatters.aap_formatter_common import map_priority
from datetime import datetime
import io


class ANPAFormatterTest(SuperdeskTestCase):
    subscribers = [{"_id": "1", "name": "notes", "subscriber_type": SUBSCRIBER_TYPES.WIRE, "media_type": "media",
                    "is_active": True, "sequence_num_settings": {"max": 10, "min": 1},
                    "destinations": [{"name": "ANPA", "delivery_type": "email", "format": "ANPA",
                                      "config": {"recipients": "test@sourcefabric.org"}
                                      }]
                    }]

    article = {
        'source': 'AAP',
        '_updated': datetime.strptime('2015-05-29 05:46', '%Y-%m-%d %H:%M'),
        'anpa_category': [{'qcode': 'a'}],
        'headline': 'This is a test headline',
        'slugline': 'slugline',
        'subject': [{'qcode': '02011001'}],
        'anpa_take_key': 'take_key',
        'urgency': 5,
        'unique_id': '1',
        'body_html': 'The story body',
        'type': 'text',
        'word_count': '1',
        'priority': 1,
        'task': {'desk': 1},
        'body_footer': 'call helpline 999 if you are planning to quit smoking'
    }

    desks = [{'_id': 1, 'name': 'National'},
             {'_id': 2, 'name': 'Sports'},
             {'_id': 3, 'name': 'Finance'}]

    def setUp(self):
        super().setUp()
        self.app.data.insert('subscribers', self.subscribers)
        self.app.data.insert('desks', self.desks)
        init_app(self.app)

    def testANPAFormatter(self):
        subscriber = self.app.data.find('subscribers', None, None)[0]

        f = AAPAnpaFormatter()
        seq, item = f.format(self.article.copy(), subscriber)[0]

        self.assertGreater(int(seq), 0)

        lines = io.StringIO(item.decode())

        line = lines.readline()
        self.assertTrue('axx' in line[1:])

        line = lines.readline()
        self.assertEqual(line[:3], '')  # Skip the sequence

        line = lines.readline()
        self.assertEqual(line[0:20], 'f a bc-slugline   ')  # skip the date

        line = lines.readline()
        self.assertEqual(line.strip(), 'This is a test headline')

        line = lines.readline()
        self.assertEqual(line.strip(), 'slugline take_key')

        line = lines.readline()
        self.assertEqual(line.strip(), 'The story bodycall helpline 999 if you are planning to quit smoking')

        line = lines.readline()
        self.assertEqual(line.strip(), 'AAP')

    def testANPAWithNoSelectorsFormatter(self):
        subscriber = self.app.data.find('subscribers', None, None)[0]
        subscriber['name'] = 'not notes'

        f = AAPAnpaFormatter()
        seq, item = f.format(self.article.copy(), subscriber)[0]

        self.assertGreater(int(seq), 0)

        lines = io.StringIO(item.decode())

        line = lines.readline()
        self.assertEqual(line[:3], '')  # Skip the sequence

        line = lines.readline()
        self.assertEqual(line[0:20], 'f a bc-slugline   ')  # skip the date

        line = lines.readline()
        self.assertEqual(line.strip(), 'This is a test headline')

        line = lines.readline()
        self.assertEqual(line.strip(), 'slugline take_key')

        line = lines.readline()
        self.assertEqual(line.strip(), 'The story bodycall helpline 999 if you are planning to quit smoking')

        line = lines.readline()
        self.assertEqual(line.strip(), 'AAP')

    def testANPAWithBylineFormatter(self):
        subscriber = self.app.data.find('subscribers', None, None)[0]
        subscriber['name'] = 'not notes'
        byline_article = dict(self.article.copy())
        byline_article['byline'] = 'Joe Blogs'

        f = AAPAnpaFormatter()
        seq, item = f.format(byline_article, subscriber)[0]

        self.assertGreater(int(seq), 0)

        lines = io.StringIO(item.decode())

        line = lines.readline()
        self.assertEqual(line[:3], '')  # Skip the sequence

        line = lines.readline()
        self.assertEqual(line[0:20], 'f a bc-slugline   ')  # skip the date

        line = lines.readline()
        self.assertEqual(line.strip(), 'This is a test headline')

        line = lines.readline()
        self.assertEqual(line.strip(), 'slugline take_key')

        line = lines.readline()
        self.assertEqual(line.strip(), 'Joe Blogs')

        line = lines.readline()
        self.assertEqual(line.strip(), 'The story bodycall helpline 999 if you are planning to quit smoking')

        line = lines.readline()
        self.assertEqual(line.strip(), 'AAP')

    def testMultipleCategoryFormatter(self):
        subscriber = self.app.data.find('subscribers', None, None)[0]
        multi_article = dict(self.article.copy())
        multi_article.pop('anpa_category')
        multi_article['anpa_category'] = [{'qcode': 'a'}, {'qcode': 'b'}]
        f = AAPAnpaFormatter()
        docs = f.format(multi_article, subscriber)
        self.assertEqual(len(docs), 2)
        cat = 'a'
        for seq, doc in docs:
            lines = io.StringIO(doc.decode())
            line = lines.readline()
            line = lines.readline()
            line = lines.readline()
            self.assertEqual(line[2:3], cat)  # skip the date
            cat = 'b'

    def test_process_headline_empty_sequence_short_headline(self):
        f = AAPAnpaFormatter()
        article = {'headline': '1234567890' * 5}
        anpa = []
        f._process_headline(anpa, article, 'a')
        self.assertEqual(anpa[0], b'12345678901234567890123456789012345678901234567890')

    def test_process_headline_empty_sequence_long_headline(self):
        f = AAPAnpaFormatter()
        article = {'headline': '1234567890' * 7}
        anpa = []
        f._process_headline(anpa, article, 'a')
        self.assertEqual(anpa[0], b'1234567890123456789012345678901234567890123456789012345678901234')

    def test_process_headline_with_sequence_short_headline(self):
        f = AAPAnpaFormatter()
        article = {'headline': '1234567890=7', 'sequence': 7}
        anpa = []
        f._process_headline(anpa, article, 'a')
        self.assertEqual(anpa[0], b'1234567890=7')

    def test_process_headline_with_sequence_long_headline(self):
        f = AAPAnpaFormatter()
        article1 = {'headline': '1234567890' * 7 + '=7', 'sequence': 7}
        anpa = []
        f._process_headline(anpa, article1, 'a')
        self.assertEqual(anpa[0], b'12345678901234567890123456789012345678901234567890123456789012=7')
        article2 = {'headline': '1234567890' * 7 + '=7', 'sequence': 17}
        anpa = []
        f._process_headline(anpa, article2, 'a')
        self.assertEqual(anpa[0], b'1234567890123456789012345678901234567890123456789012345678901=17')

    def test_process_headline_locator_inject(self):
        f = AAPAnpaFormatter()
        article3 = {'headline': '1234567890' * 3, 'place': [{'qcode': 'VIC', 'name': 'VIC'}]}
        anpa = []
        f._process_headline(anpa, article3, 'a')
        self.assertEqual(anpa[0], b'VIC:123456789012345678901234567890')

    def test_map_priority(self):
        self.assertEqual('f', map_priority(1))
        self.assertEqual('u', map_priority(2))
        self.assertEqual('b', map_priority(3))
        self.assertEqual('r', map_priority(4))
        self.assertEqual('r', map_priority(5))
        self.assertEqual('r', map_priority(6))
        self.assertEqual('r', map_priority(None))
        self.assertEqual('r', map_priority(7))
        self.assertEqual('r', map_priority(''))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
import os


def _load_json(file_path):
    """
    Reads JSON string from the file located in file_path.

    :param file_path: path of the file having JSON string.
    :return: JSON Object
    """
    with open(file_path, 'r', encoding='utf-8') as f:
        return json.load(f)


_dir_name = os.path.dirname(os.path.realpath(__file__))
_locators_file_path = os.path.join(_dir_name, 'data', 'locators.json')
locators = _load_json(_locators_file_path)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
import logging
from copy import deepcopy
from superdesk import get_resource_service
from superdesk.errors import SuperdeskApiError
from superdesk.media.media_operations import crop_image, process_file_from_stream
from superdesk.upload import url_for_media
from superdesk.metadata.item import CONTENT_TYPE, ITEM_TYPE


logger = logging.getLogger(__name__)


class CropService():

    crop_sizes = []

    def validate_crop(self, original, crop_name, doc):
        """

        :param dict original: original item
        :param str crop_name: name of the crop
        :param dict doc: crop co-ordinates
        :raises SuperdeskApiError.badRequestError:
            For following conditions:
            1) if type != picture
            2) if renditions are missing in the original image
            3) if original rendition is missing
            4) Crop name is invalid
        """
        # Check if type is picture
        if original[ITEM_TYPE] != CONTENT_TYPE.PICTURE:
            raise SuperdeskApiError.badRequestError(message='Only images can be cropped!')

        # Check if the renditions exists
        if not original.get('renditions'):
            raise SuperdeskApiError.badRequestError(message='Missing renditions!')

        # Check if the original rendition exists
        if not original.get('renditions').get('original'):
            raise SuperdeskApiError.badRequestError(message='Missing original rendition!')

        # Check if the crop name is valid
        crop = self.get_crop_by_name(crop_name)
        if not crop and 'CropLeft' in doc:
            raise SuperdeskApiError.badRequestError(message='Unknown crop name! (name=%s)' % crop_name)

        self._validate_aspect_ratio(crop, doc)

    def _validate_aspect_ratio(self, crop, doc):
        """
        Checks if the aspect ratio is consistent with one in defined in spec
        :param crop: Spec parameters
        :param doc: Posted parameters
        :raises SuperdeskApiError.badRequestError:
        """
        if 'CropLeft' not in doc:
            return

        width = doc['CropRight'] - doc['CropLeft']
        height = doc['CropBottom'] - doc['CropTop']

        if width < crop['width'] or height < crop['height']:
            raise SuperdeskApiError.badRequestError(
                message='Wrong crop size. Minimum crop size is {}x{}.'.format(crop['width'], crop['height']))

        doc_ratio = round(width / height, 1)
        spec_ratio = round(crop['width'] / crop['height'], 1)
        if doc_ratio != spec_ratio:
            raise SuperdeskApiError.badRequestError(message='Wrong aspect ratio!')

    def get_crop_by_name(self, crop_name):
        """
        Finds the crop in the list of crops by name
        :param crop_name: Crop name
        :return: Matching crop or None
        """
        if not self.crop_sizes:
            self.crop_sizes = get_resource_service('vocabularies').find_one(req=None, _id='crop_sizes').get('items')

        if not self.crop_sizes:
            raise SuperdeskApiError.badRequestError(message='Crops sizes couldn\'t be loaded!')

        return next((c for c in self.crop_sizes if c.get('name', '').lower() == crop_name.lower()), None)

    def create_crop(self, original, crop_name, crop_data):
        """
        Create a new crop based on the crop co-ordinates
        :param original: Article to add the crop
        :param crop_name: Name of the crop
        :param doc: Crop details
        :raises SuperdeskApiError.badRequestError
        :return dict: modified renditions
        """
        renditions = original.get('renditions', {})
        original_crop = renditions.get(crop_name, {})
        fields = ('CropLeft', 'CropTop', 'CropRight', 'CropBottom')
        crop_created = False

        if any(crop_data.get(name) != original_crop.get(name) for name in fields):
            original_image = renditions.get('original', {})
            original_file = superdesk.app.media.fetch_rendition(original_image)
            if not original_file:
                raise SuperdeskApiError.badRequestError('Original file couldn\'t be found')

            try:
                cropped, out = crop_image(original_file, crop_name, crop_data)

                if not cropped:
                    raise SuperdeskApiError.badRequestError('Saving crop failed.')

                renditions[crop_name] = self._save_cropped_image(out, original_image, crop_data)
                crop_created = True
            except SuperdeskApiError:
                raise
            except Exception as ex:
                raise SuperdeskApiError.badRequestError('Generating crop failed: {}'.format(str(ex)))

        return renditions, crop_created

    def _save_cropped_image(self, file_stream, original, doc):
        """
        Saves the cropped image and returns the crop dictionary
        :param file_stream: cropped image stream
        :param original: original rendition
        :param doc: crop data
        :return dict: Crop values
        :raises SuperdeskApiError.internalError
        """
        crop = {}
        try:
            file_name, content_type, metadata = process_file_from_stream(file_stream,
                                                                         content_type=original.get('mimetype'))
            file_stream.seek(0)
            file_id = superdesk.app.media.put(file_stream, filename=file_name,
                                              content_type=content_type,
                                              resource='upload',
                                              metadata=metadata)
            crop['media'] = file_id
            crop['mimetype'] = content_type
            crop['href'] = url_for_media(file_id)
            crop['CropTop'] = doc.get('CropTop', None)
            crop['CropLeft'] = doc.get('CropLeft', None)
            crop['CropRight'] = doc.get('CropRight', None)
            crop['CropBottom'] = doc.get('CropBottom', None)
            return crop
        except Exception as ex:
            try:
                superdesk.app.media.delete(file_id)
            except:
                pass
            raise SuperdeskApiError.internalError('Generating crop failed: {}'.format(str(ex)))

    def _delete_crop_file(self, file_id):
        """
        Delete the crop file
        :param Object_id file_id: Object_Id of the file.
        """
        try:
            superdesk.app.media.delete(file_id)
        except:
            logger.exception("Crop File cannot be deleted. File_Id {}".format(file_id))

    def create_multiple_crops(self, updates, original):
        """
        Create multiple crops based on the renditions.
        :param dict updates: update item
        :param dict original: original of the updated item
        """
        update_renditions = updates.get('renditions', {})
        if original.get(ITEM_TYPE) == CONTENT_TYPE.PICTURE and update_renditions:
            renditions = original.get('renditions', {})
            original_copy = deepcopy(original)
            for key in update_renditions:
                if self.get_crop_by_name(key):
                    renditions, crop_created = self.create_crop(original_copy, key, update_renditions.get(key, {}))

            updates['renditions'] = renditions

    def validate_multiple_crops(self, updates, original):
        """
        Validate crops for the image
        :param dict updates: update item
        :param dict original: original of the updated item
        """
        renditions = updates.get('renditions', {})
        if renditions and original.get(ITEM_TYPE) == CONTENT_TYPE.PICTURE:
            for key in renditions:
                self.validate_crop(original, key, renditions.get(key, {}))

    def delete_replaced_crop_files(self, updates, original):
        """
        Delete the replaced crop files.
        :param dict updates: update item
        :param dict original: original of the updated item
        """
        update_renditions = updates.get('renditions', {})
        if original.get(ITEM_TYPE) == CONTENT_TYPE.PICTURE and update_renditions:
            renditions = original.get('renditions', {})
            for key in update_renditions:
                if self.get_crop_by_name(key) and \
                        update_renditions.get(key, {}).get('media') != \
                        renditions.get(key, {}).get('media'):
                    self._delete_crop_file(renditions.get(key, {}).get('media'))



from .common import BasePublishService, BasePublishResource, ITEM_CORRECT
from eve.utils import config
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE, EMBARGO
from superdesk.metadata.packages import PACKAGE_TYPE
from superdesk.publish import SUBSCRIBER_TYPES
from superdesk import get_resource_service
from superdesk.utc import utcnow
from apps.archive.common import set_sign_off, ITEM_OPERATION, insert_into_versions
from superdesk.media.crop import CropService
import logging

logger = logging.getLogger(__name__)


class CorrectPublishResource(BasePublishResource):

    def __init__(self, endpoint_name, app, service):
        super().__init__(endpoint_name, app, service, 'correct')


class CorrectPublishService(BasePublishService):
    publish_type = 'correct'
    published_state = 'corrected'

    def on_update(self, updates, original):
        CropService().validate_multiple_crops(updates, original)
        super().on_update(updates, original)
        updates[ITEM_OPERATION] = ITEM_CORRECT
        set_sign_off(updates, original)

    def on_updated(self, updates, original):
        """
        Locates the published or corrected non-take packages containing the corrected item
        and corrects them
        :param updates: correction
        :param original: original story
        """
        original_updates = dict()
        original_updates['operation'] = updates['operation']
        original_updates[ITEM_STATE] = updates[ITEM_STATE]
        super().on_updated(updates, original)
        CropService().delete_replaced_crop_files(updates, original)
        packages = self.package_service.get_packages(original[config.ID_FIELD])
        if packages and packages.count() > 0:
            archive_correct = get_resource_service('archive_correct')
            processed_packages = []
            for package in packages:
                if package[ITEM_STATE] in [CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED] and \
                        package.get(PACKAGE_TYPE, '') == '' and \
                        str(package[config.ID_FIELD]) not in processed_packages:
                    original_updates['groups'] = package['groups']

                    if updates.get('headline'):
                        self.package_service.update_field_in_package(original_updates, original[config.ID_FIELD],
                                                                     'headline', updates.get('headline'))

                    if updates.get('slugline'):
                        self.package_service.update_field_in_package(original_updates, original[config.ID_FIELD],
                                                                     'slugline', updates.get('slugline'))

                    archive_correct.patch(id=package[config.ID_FIELD], updates=original_updates)
                    insert_into_versions(id_=package[config.ID_FIELD])
                    processed_packages.append(package[config.ID_FIELD])

    def update(self, id, updates, original):
        CropService().create_multiple_crops(updates, original)
        super().update(id, updates, original)

    def get_subscribers(self, doc, target_media_type):
        """
        Get the subscribers for this document based on the target_media_type for article Correction.
        1. The article is sent to Subscribers (digital and wire) who has received the article previously.
        2. For subsequent takes, only published to previously published wire clients. Digital clients don't get
           individual takes but digital client takes package.
        3. If the item has embargo and is a future date then fetch active Wire Subscribers.
           Otherwise fetch Active Subscribers. After fetching exclude those who received the article previously from
           active subscribers list.
        4. If article has 'targeted_for' property then exclude subscribers of type Internet from Subscribers list.
        5. Filter the subscriber that have not received the article previously against publish filters
        and global filters for this document.
        :param doc: Document to correct
        :param target_media_type: dictate if the doc being queued is a Takes Package or an Individual Article.
                Valid values are - Wire, Digital. If Digital then the doc being queued is a Takes Package and if Wire
                then the doc being queues is an Individual Article.
        :return: (list, list) List of filtered subscribers, List of subscribers that have not received item previously
        """
        subscribers, subscribers_yet_to_receive = [], []
        # step 1
        query = {'$and': [{'item_id': doc[config.ID_FIELD]},
                          {'publishing_action': {'$in': [CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED]}}]}

        subscribers = self._get_subscribers_for_previously_sent_items(query)

        if subscribers:
            # step 2
            if not self.takes_package_service.get_take_package_id(doc):
                # Step 3
                query = {'is_active': True}
                if doc.get(EMBARGO) and doc.get(EMBARGO) > utcnow():
                    query['subscriber_type'] = SUBSCRIBER_TYPES.WIRE

                active_subscribers = list(get_resource_service('subscribers').get(req=None, lookup=query))
                subscribers_yet_to_receive = [a for a in active_subscribers
                                              if not any(a[config.ID_FIELD] == s[config.ID_FIELD]
                                                         for s in subscribers)]

            if len(subscribers_yet_to_receive) > 0:
                # Step 4
                if doc.get('targeted_for'):
                    subscribers_yet_to_receive = list(self.non_digital(subscribers_yet_to_receive))
                # Step 5
                subscribers_yet_to_receive = \
                    self.filter_subscribers(doc, subscribers_yet_to_receive,
                                            SUBSCRIBER_TYPES.WIRE if doc.get('targeted_for') else target_media_type)

        return subscribers, subscribers_yet_to_receive



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from superdesk.utils import get_random_string
from elasticsearch.helpers import reindex
from eve_elastic import get_es, get_indices
import elasticsearch
import superdesk


class RebuildElasticIndex(superdesk.Command):
    """
    Rebuild the elastic index from existing data by creating a new index with
    the same alias as the configured index, puts the new mapping and delete the old index.
    """
    def run(self):
        index_name = superdesk.app.config['ELASTICSEARCH_INDEX']
        print('Starting index rebuilding for index: ', index_name)
        try:
            es = get_es(superdesk.app.config['ELASTICSEARCH_URL'])
            clone_name = index_name + '-' + get_random_string()
            print('Creating index: ', clone_name)
            superdesk.app.data.elastic.create_index(clone_name, superdesk.app.config['ELASTICSEARCH_SETTINGS'])
            print('Putting mapping for index: ', clone_name)
            superdesk.app.data.elastic.put_mapping(superdesk.app, clone_name)
            print('Starting index rebuilding.')
            reindex(es, index_name, clone_name)
            print('Finished index rebuilding.')
            print('Deleting index: ', index_name)
            get_indices(es).delete(index_name)
            print('Creating alias: ', index_name)
            get_indices(es).put_alias(index=clone_name, name=index_name)
            print('Alias created.')
        except elasticsearch.exceptions.NotFoundError as nfe:
            print(nfe)
        print('Index {0} rebuilt successfully.'.format(index_name))

superdesk.command('app:rebuild_elastic_index', RebuildElasticIndex())



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.publish.publish_service import PublishService
from superdesk.publish import register_transmitter
from superdesk.errors import PublishFileError

errors = [PublishFileError.fileSaveError().get_error_description()]


class FilePublishService(PublishService):
    def _transmit(self, queue_item, subscriber):
        config = queue_item.get('destination', {}).get('config', {})

        try:
            if isinstance(queue_item['formatted_item'], bytes):
                self.copy_file(config, queue_item)
            elif isinstance(queue_item['formatted_item'], str):
                queue_item['formatted_item'] = queue_item['formatted_item'].encode('utf-8')
                self.copy_file(config, queue_item)
            else:
                raise Exception
        except Exception as ex:
            raise PublishFileError.fileSaveError(ex, config)

    def copy_file(self, config, queue_item):
        with open('{}/{}-{}-{}.{}'.format(config['file_path'],
                                          queue_item['item_id'].replace(':', '-'),
                                          str(queue_item.get('item_version', '')),
                                          str(queue_item.get('published_seq_num', '')),
                                          config.get('file_extension', 'txt')), 'wb') as f:
            f.write(queue_item['formatted_item'])


register_transmitter('File', FilePublishService(), errors)



"""Here you can import/implement macros.

This module will get reloaded per request to be up to date.

Use `superdesk.macro_register.macros.register` for registration.
"""

import os
import sys
import imp
import importlib


if os.path.isdir('macros'):
    macros = [f[:-3] for f in os.listdir(os.path.abspath('macros'))
              if f.endswith('.py') and not f.endswith('_test.py') and not f.startswith('__')]

    for macro in macros:
        try:
            module = 'macros.{}'.format(macro)
            if module in sys.modules.keys():
                m = sys.modules[module]
                imp.reload(m)
            else:
                importlib.import_module(module)
        except:
            pass



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

"""Superdesk push notifications"""

import os
import logging
import asyncio
import websockets

from flask import current_app as app, json
from datetime import datetime
from superdesk.utils import json_serialize_datetime_objectId

logger = logging.getLogger(__name__)


class ClosedSocket():
    """Mimic closed socket to simplify logic when connection
    can't be estabilished at first place.
    """
    def __init__(self):
        self.open = False

    def close(self):
        pass


def init_app(app):
    """Create websocket connection and put it on app object."""
    host = app.config['WS_HOST']
    port = app.config['WS_PORT']
    try:
        try:
            loop = asyncio.get_event_loop()
        except:
            loop = asyncio.new_event_loop()

        app.notification_client = loop.run_until_complete(websockets.connect('ws://%s:%s/server' % (host, port)))
        logger.info('websocket connected on=%s:%s' % app.notification_client.local_address)
    except (RuntimeError, OSError):
        # not working now, but we can try later when actually sending something
        app.notification_client = ClosedSocket()


def _notify(**kwargs):
    """Send out all kwargs as json string."""
    kwargs.setdefault('_created', datetime.utcnow().isoformat())
    kwargs.setdefault('_process', os.getpid())
    message = json.dumps(kwargs, default=json_serialize_datetime_objectId)

    @asyncio.coroutine
    def send_message():
        yield from app.notification_client.send(message)

    try:
        loop = asyncio.get_event_loop()
    except:
        loop = asyncio.new_event_loop()
    loop.run_until_complete(send_message())


def push_notification(name, **kwargs):
    """Push notification to websocket.

    In case socket is closed it will try to reconnect.

    :param name: event name
    """
    logger.info('pushing event {0} ({1})'.format(name, json.dumps(kwargs, default=json_serialize_datetime_objectId)))

    if not app.notification_client.open:
        app.notification_client.close()
        init_app(app)

    if not app.notification_client.open:
        logger.info('No connection to websocket server. Dropping event %s' % name)
        return

    try:
        _notify(event=name, extra=kwargs)
    except Exception as err:
        logger.exception(err)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from test_factory import SuperdeskTestCase
from apps.publish import init_app
from apps.publish.formatters.field_mappers.selectorcode_mapper import SelectorcodeMapper


class SelectorcodeMapperTest(SuperdeskTestCase):

    desks = [{'_id': 1, 'name': 'National'},
             {'_id': 2, 'name': 'Sport'},
             {'_id': 3, 'name': 'Finance'}]

    def setUp(self):
        super().setUp()
        self.app.data.insert('desks', self.desks)
        init_app(self.app)

    def test_is_in_subject(self):
        article = {
            'subject': [{'qcode': '04001005'}, {'qcode': '15011002'}],
        }
        f = SelectorcodeMapper()
        self.assertTrue(f._is_in_subject(article, '150'))
        self.assertFalse(f._is_in_subject(article, '151'))
        self.assertTrue(f._is_in_subject(article, '04001'))

    def test_join_selector_codes(self):
        f = SelectorcodeMapper()
        result = f._join_selector_codes('ipnewS', 'newsi', 'cnewsi', 'cnewsi')
        result_list = result.split()
        self.assertEqual(len(result_list), 12)

    def test_set_selector_codes(self):
        article = {
            'task': {'desk': 1},
            'slugline': 'Test',
            'urgency': 3
        }
        subscriber = {'name': 'ipnews'}
        f = SelectorcodeMapper()
        odbc_item = {}
        with self.app.app_context():
            f.map(article, 'A', subscriber=subscriber, formatted_item=odbc_item)
            self.assertSetEqual(set(odbc_item['selector_codes'].split()),
                                set('and axd pnd cxd 0fh 0ir 0px 0ah 0hw cxx axx cnd 0nl az pxd pxx'.split()))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from test_factory import SuperdeskTestCase
import datetime
from .noise11_derive_metadata import noise11_derive_metadata


class Noise11DeriveMetaDataTests(SuperdeskTestCase):

    vocab = [{'_id': 'categories', 'items': [{'is_active': True, 'name': 'Entertainment', 'qcode': 'e'}]}]
    now = datetime.datetime(2015, 6, 13, 11, 45, 19, 0)

    def setUp(self):
        super().setUp()
        self.app.data.insert('vocabularies', self.vocab)

    def simple_case_test(self):
        item = dict()
        item['firstcreated'] = datetime.datetime(2015, 10, 26, 11, 45, 19, 0)
        item['source'] = 'NOISE11'
        noise11_derive_metadata(item)
        self.assertEqual(item['dateline']['text'], 'SYDNEY Oct 26 NOISE11 -')



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
from superdesk.resource import Resource
from superdesk.utils import required_string


class AAPMMResource(superdesk.Resource):
    resource_methods = ['GET', 'POST']
    schema = {
        'guid': required_string,
        'desk': Resource.rel('desks', False, nullable=True)
    }
